﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
//using System.Ling;
using System.Text;
using System.Windows.Forms;
using System.Threading;

using System.IO;    //File operations
using System.Collections; //HashTable
using System.Security.Permissions;
using System.Diagnostics;
using System.Xml;
using weka;
using IKVM;
using libsvm;

//tree namespace
using Common;


/*ZedGraph library used under License-LGPL
 http://sourceforge.net/projects/zedgraph
*/
using ZedGraph; //for Graph


[assembly: PermissionSetAttribute(SecurityAction.RequestMinimum, Name = "Internet")]
[assembly: PermissionSetAttribute(SecurityAction.RequestOptional, Unrestricted = true)]

namespace WekaAnonymize
{
    public partial class frmWekaAnonymize : Form
    {
        StreamReader dataFileStreamReader = null;
        StreamReader namesFileStreamReader = null;
        StreamWriter generalizedDataFileStreamWriter = null;
        StreamWriter logValues = null;
        
        //XML File Variables
        XmlDocument xDoc;

        XmlNodeList k, classes, qids, sids, inputFile, inputLines, namesFile, outputFile, KullbackLeiblerDivergence, NormalisedMutualInformation;
        XmlNodeList trainingRatio, testRatio, testFolds;
        XmlNodeList crossValidationFolds;

        String[] classifierFlag;

        const int J48 = 0, NAIVE_BAYES = 1, LOGISTICS = 2, SVM = 3;
        String[] trainingTuples, testTuples;
        bool trainingFlag, testFlag;
        ArrayList[] J48_foldAccuracyValues, NaiveBayes_foldAccuracyValues, Logistics_foldAccuracyValues, SVM_foldAccuracyValues;

        String[] qiNames;
        
        TreeView[] QI_GeneralizationTree;
        
        QI[] QIs;
        ArrayList[] QIs_Levels;
        Level[] generalized_QIs_Levels;
        Level[] generalized_QIs_Suppressed_Levels;
        ArrayList [] suppressedClasses_QIs;
        String[] tuples;
        int[] sensitiveFlag;
        int[] maximumTradeoffLevels;

        ArrayList generalizedTuples;
        ArrayList generalizedDistinctTuples;

        int totalQIs = 0;
        
        //sorting variables
        DistinctTuple[] distinct_tuple_objects;

        int[] class_count;
        int[] class_count_non_sorted;
        Hashtable class_count_tuples;
        Hashtable class_high_low;
        Hashtable class_count_tuples_suppressed;

        DateTime dateTimeStart;
        long startMilliSeconds;
        Process currentProcess;


        public frmWekaAnonymize(String [] args)
        {
            InitializeComponent();

            btnSelectFile_Click(args);

        }//public frmWekaAnonymize(String [] args)

        private void btnSelectFile_Click(object sender, EventArgs e)
        {
            String[] args = new String[1];
            btnSelectFile_Click(args);
        }
        private void btnSelectFile_Click(String [] args)
        {
            openFileDialogData.Filter = "XML Input Files (*.xml)|*.xml";
            openFileDialogData.Multiselect = false;

            DialogResult openFileDialogResult = openFileDialogData.ShowDialog();

            if (openFileDialogResult == DialogResult.Cancel)
                return;

            trainingFlag = false;
            testFlag = false;

            String path = openFileDialogData.FileName;
            path = path.Substring(0, path.LastIndexOf("\\"));

            Directory.CreateDirectory(path + "\\output");


                       
            //read XML file
            xDoc = new XmlDocument();
            xDoc.Load(openFileDialogData.FileName);
            //xDoc.Load(args[0]);

            k = xDoc.GetElementsByTagName("k");
            classes = xDoc.GetElementsByTagName("classes");
            qids = xDoc.GetElementsByTagName("attribute");
            sids = xDoc.GetElementsByTagName("sensitive");
            inputFile = xDoc.GetElementsByTagName("inputfile");
            inputLines = xDoc.GetElementsByTagName("lines");
            namesFile = xDoc.GetElementsByTagName("namesfile");
            outputFile = xDoc.GetElementsByTagName("outputfile");
            KullbackLeiblerDivergence = xDoc.GetElementsByTagName("KullbackLeiblerDivergence");
            NormalisedMutualInformation = xDoc.GetElementsByTagName("NormalisedMutualInformation");

            totalQIs = qids.Count;

            //MessageBox.Show(
            //  "k: " + k[0].InnerText + "\n" +
            //  "classes: " + classes[0].InnerText + "\n" +
            //  "qid: " + qids.Count + "\n" +
            //  "qid: " + qids[1].InnerText + "\n" +
            //  "sid: " + sids[0].InnerText + "\n" +
            //  "inputFile: " + inputFile[0].InnerText + "\n" +
            //  "namesFile: " + namesFile[0].InnerText + "\n" +
            //  "outputFile: " + outputFile[0].InnerText + "\n"
            //  );


            qiNames = new String[totalQIs];
            for (int a = 0; a < totalQIs; a++)
              qiNames[a] = qids[a].InnerText;  //get node list haveing the name of QI, there will ALWAYS be one node having name of QI

          sensitiveFlag = new int[sids.Count];
          for (int b = 0; b < sids.Count; b++)
              sensitiveFlag[b] = int.Parse(sids[b].InnerText);

          QI_GeneralizationTree = new  TreeView[totalQIs];
          QIs = new QI[totalQIs];
          QIs_Levels = new ArrayList[totalQIs];
            
          //lblFilePath.Text = openFileDialogData.SafeFileName;         //show file path in dialog
          InitializeDataFileStream(true);
          InitializeNamesFileStream();  //use this file when read level information from .names file
            
          logValues = new StreamWriter("disnt_XML.txt");
          logValues.AutoFlush = true;

          dateTimeStart = DateTime.Now;

          logValues.WriteLine("File Reading Starts at " +
                                                            dateTimeStart.Day.ToString() + "/" +
                                                            dateTimeStart.Month.ToString() + "/" +
                                                            dateTimeStart.Year.ToString() + "/ " +
                                                            dateTimeStart.Hour.ToString() + ":" +
                                                            dateTimeStart.Minute.ToString() + ":" +
                                                            dateTimeStart.Second.ToString() + ":" +
                                                            dateTimeStart.Millisecond.ToString());

            currentProcess = Process.GetCurrentProcess();
            startMilliSeconds = (long)currentProcess.TotalProcessorTime.TotalMilliseconds;


            for (int a = 0; a < totalQIs; a++)
            {
                QIs_Levels[a] = new ArrayList(QIs[a].levels);
                CreateLevels(ref QIs_Levels[a], a);
            }

           
            logValues.WriteLine("");

            DateTime dateTimeEnd;
            dateTimeEnd = DateTime.Now;


            logValues.WriteLine("File reading ends after creating levels for each attribute at " +
                                                            dateTimeEnd.Day.ToString() + "/" +
                                                            dateTimeEnd.Month.ToString() + "/" +
                                                            dateTimeEnd.Year.ToString() + "/ " +
                                                            dateTimeEnd.Hour.ToString() + ":" +
                                                            dateTimeEnd.Minute.ToString() + ":" +
                                                            dateTimeEnd.Second.ToString() + ":" +
                                                            dateTimeEnd.Millisecond.ToString());

            logValues.WriteLine("Total time taken in reading and levels creating " + (dateTimeEnd - dateTimeStart).ToString());

            currentProcess = Process.GetCurrentProcess();
            long endMilliSeconds = (long)currentProcess.TotalProcessorTime.TotalMilliseconds;
            logValues.WriteLine("Total CPU time taken in reading and levels creating " + (endMilliSeconds - startMilliSeconds).ToString());


            logValues.WriteLine("");

            //
            //write origional data into weka file format
            //
            CreateOrigionalWekaDataSet();


            //
            //generalize and write generalized data into weka file format
            //
            GeneralizeDataToToMaximumTradeoffLevel(QIs_Levels);

            //count distict tuples after generalization
            CountDistinctTuples(ref generalizedTuples, ref generalizedDistinctTuples);


            //sort tuples first number of distinct tuples and secondly total number of distinct tuples
            SortGeneralizedTuples();

            logValues.WriteLine("");
            logValues.WriteLine("Number of Distinct Tuples: " + generalizedDistinctTuples.Count.ToString());
            logValues.WriteLine("");
            
            logValues.WriteLine("Total Equivalence Classes: " + class_count_tuples.Count.ToString());

            int suppressClasses = int.Parse(classes[0].InnerText);
            
            if (suppressClasses > class_count_tuples.Count)
            {
               // MessageBox.Show("Maximum Equivalence Classes available for suppression: " + class_count_tuples.Count + "\ninstead of input parameter in XML file: " + suppressClasses);
                suppressClasses = class_count_tuples.Count;
            }

            if(suppressClasses > 0)
                suppressClasses = SuppressClasses(suppressClasses);
            
            logValues.WriteLine("");

            ArrayList tempSuppressed;
            SuppressedClass tempSuppressedClass;

            for (int b = 0; b < totalQIs; b++)
            {
                if (sensitiveFlag[b] == 1)
                    continue;
                else
                    for (int a = 1; a <= suppressClasses; a++)
                    {
                        tempSuppressed = (ArrayList)suppressedClasses_QIs[b];
                        tempSuppressedClass = (SuppressedClass)tempSuppressed[a];
                        logValues.WriteLine(qiNames[b] + " Distribution Distance After Suppressing Class " + a.ToString() + " : " + tempSuppressedClass.distribution_distance.ToString());
                    }
                logValues.WriteLine("");
            }


            logValues.WriteLine("==Normalized Distribution Distance==");
            for (int b = 0; b < totalQIs; b++)
            {
                logValues.WriteLine(qiNames[b]);
                if (sensitiveFlag[b] == 1)
                    continue;
                else
                    for (int a = 1; a <= suppressClasses; a++)
                    {
                        tempSuppressed = (ArrayList)suppressedClasses_QIs[b];
                        tempSuppressedClass = (SuppressedClass)tempSuppressed[a];
                        
                        logValues.WriteLine(qiNames[b] + " Normalized Distribution Distance After Suppressing Class " + a.ToString() + " : " + tempSuppressedClass.distribution_distance_normalized.ToString());
                    }
                logValues.WriteLine("");
            }

        logValues.WriteLine("==Information Gain Ratio==");
            for (int b = 0; b < totalQIs; b++)
            {
                logValues.WriteLine(qiNames[b]);

                if (sensitiveFlag[b] == 1)
                    continue;
                else
                    for (int a = 1; a <= suppressClasses; a++)
                    {
                        tempSuppressed = (ArrayList)suppressedClasses_QIs[b];
                        tempSuppressedClass = (SuppressedClass)tempSuppressed[a];
                        
                        logValues.WriteLine(qiNames[b] + " Information Gain Ratio After Suppressing Class " + a.ToString() + " : " + tempSuppressedClass.info_gain_ratio.ToString());
            
                    }
                    

                logValues.WriteLine("");
            }

            
            //GRAPH DRAWING STARTS
            /*ZedGraph library used under License-LGPL
             * 
             http://sourceforge.net/projects/zedgraph
             */
            // Setup the graph

            CreateGraphAGain(zedGraphGain);
            CreateGraphTradeoff(zedGraphTradeoff);
            CreateGraphEntropy(zedGraphEntropy);
            CreateGraphHistogram(zedGraphHistogram);
            CreateGraphClassFrequency(zedGraphClassFrequency);
            CreateGraphDistributionDistance(zedGraphDistance);
            CreateGraphInfoGainRatio(zedGraphInfoGainRatio);
            CreateGraphClassSizeXfrequency(zedGraphClassSizeXFrequency);
            CreateGraphASpecLoss(zedGraphSpecLoss);
            CreateGraphSuppressionCount(zedGraphTotalSuppressionCount);

            ///////////////////////////////////////////////////////////////////////
            //Calculation of accuracy values using weka APIs

            XmlNodeList wekaClassifiers = xDoc.GetElementsByTagName("wekaClassifiers");

            if (wekaClassifiers.Count == 0)
            {
                logValues.Flush();
                logValues.Close();
                MessageBox.Show("done");
                return;
            }

            return;

            classifierFlag = new String[4];

            for (int a = 0; a < classifierFlag.Length; a++)
            {
                switch (a)
                {
                    case J48:
                        wekaClassifiers = xDoc.GetElementsByTagName("J48");

                        if (wekaClassifiers.Count == 1)
                            classifierFlag[J48] = wekaClassifiers[0].InnerText;
                        else
                            classifierFlag[J48] = "false";
                        break;

                    case NAIVE_BAYES:
                        wekaClassifiers = xDoc.GetElementsByTagName("NaiveBayes");

                        if (wekaClassifiers.Count == 1)
                            classifierFlag[NAIVE_BAYES] = wekaClassifiers[0].InnerText;
                        else
                            classifierFlag[NAIVE_BAYES] = "false";
                        break;

                    case LOGISTICS:
                        wekaClassifiers = xDoc.GetElementsByTagName("Logistics");

                        if (wekaClassifiers.Count == 1)
                            classifierFlag[LOGISTICS] = wekaClassifiers[0].InnerText;
                        else
                            classifierFlag[LOGISTICS] = "false";
                        break;

                    case SVM:
                        wekaClassifiers = xDoc.GetElementsByTagName("SVM");

                        if (wekaClassifiers.Count == 1)
                            classifierFlag[SVM] = wekaClassifiers[0].InnerText;
                        else
                            classifierFlag[SVM] = "false";
                        break;
            
                }//switch(a)

            }//for (int a = 0; a < classifierFlag.Length; a++)


            Directory.CreateDirectory(path + "\\output" + "\\temp");

            XmlNodeList suppliedTrainingTest = xDoc.GetElementsByTagName("SuppliedTrainingTestRatio");

            if (suppliedTrainingTest.Count == 1)
            {
                trainingRatio = xDoc.GetElementsByTagName("training");
                testRatio = xDoc.GetElementsByTagName("test");
                testFolds = xDoc.GetElementsByTagName("folds");

                trainingTuples = new String[(tuples.Length * int.Parse(trainingRatio[0].InnerText))/100];
                testTuples = new String[(tuples.Length * int.Parse(testRatio[0].InnerText)) / 100];

                CalculateTrainingTestWekaClassifierAccuracyValues();

            }//if (suppliedTrainingTest.Count > 1)

            XmlNodeList crossValidation = xDoc.GetElementsByTagName("CrossValidation");

            if (crossValidation.Count == 1)
            {
                crossValidationFolds = xDoc.GetElementsByTagName("CrossValidationFolds");

                CalculateCrossValidationWekaClassifierAccuracyValues();
            }

            
        
            logValues.Close();

            //if (Directory.Exists(path + "\\output\\temp"))
            //{

            //    foreach (String tempFile in Directory.GetFiles(path + "\\output\\temp"))
            //        //File.Delete(tempFile);

            //    Directory.Delete(path + "\\output\\temp");
            //}
            

            MessageBox.Show("done");


        }//private void btnSelectFile_Click(object sender, EventArgs e)

        
        //Common Structures
        public struct Group
        {
            public bool groupType;  //NUMERICAL = 0, TEXT = 1
            public String groupName;
            public double entropy;

            public int groupMinValue;
            public int groupMaxValue;

            public Int64 groupCount;

            public ArrayList classifierNames;
            public ArrayList classifierCount;

            public TreeNode groupTree;
            public ArrayList nodes;

        }//public struct Group

        public struct Level
        {
            public ArrayList levelGroups;
            public double levelEntropy;
            public double levelGain;
            public double levelSplittingEntropy;
            public double levelInfoGainRatio;
            public double levelSpecLoss;
            public double levelMutualInformation;
            public double levelMutualInformationNormalized;
            public double levelClassGain;
            public double levelTradeoff;
            public double targetAttributeEntropy;
            public int maxTradeoffLevel;
            public double maxLevelDifferenceMutualInformationNormalized;

        }//public struct Level

        public struct DistinctTuple
        {
            public ArrayList tuple;
            public Int32 count;
            public String QIs;
        }

        public struct SuppressedClass
        {
            public double distribution_distance;
            public double distribution_distance_normalized;

            public double info_gain_ratio;
        }

        public struct QI
        {
            public int levels;
        }

        //////////////////////////////////
        //weka classifier calculation methods

        //Cross Validation calculations
        public void CalculateCrossValidationWekaClassifierAccuracyValues()
        {
            
            

        }//public void CalculateCrossValidationWekaClassifierAccuracyValues()


        //Training Test folds calculations
        public void CalculateTrainingTestWekaClassifierAccuracyValues()
        {
            int folds = int.Parse(testFolds[0].InnerText);
            int testIndex = 0, trainingIndex = 0, testCount = 0;
            int suppressClasses = 0, minSuppressClasses = 0;

            for (int b = 0; b < classifierFlag.Length; b++)
            {
                switch (b)
                {
                    case J48:           J48_foldAccuracyValues = new ArrayList[folds]; break;
                    case NAIVE_BAYES: NaiveBayes_foldAccuracyValues = new ArrayList[folds]; break;
                    case LOGISTICS: Logistics_foldAccuracyValues = new ArrayList[folds]; break;
                    case SVM: SVM_foldAccuracyValues = new ArrayList[folds]; break;

                }//switch (b)

            }//for (int b = 0; b < classifierFlag.Length; b++)

            testCount = testTuples.Length;

            double J48_accuracy = 0, NaiveBayes_accuracy = 0, Logistics_accuracy = 0, LibSVM_accuracy = 0;

            XmlNodeList k_interval = xDoc.GetElementsByTagName("KInterval");
            int kInterval = int.Parse(k_interval[0].InnerText);

            StreamWriter testDataset,trainingDataset;
                                    
            for (int a = 0; a < folds; a++)
            {
                testDataset = new StreamWriter("output\\test_" + (a+1) + ".csv");
                testDataset.AutoFlush = true;

                //copy test dataset
                for (int b = 0; b < testTuples.Length; b++)
                {
                    testTuples[b] = tuples[testIndex++];
                    testDataset.Write(testTuples[b]);
                    testDataset.Write("\n");
                }

                testDataset.Flush();
                testDataset.Close();

                trainingDataset = new StreamWriter("output\\training_" + (a + 1) + ".csv");
                trainingDataset.AutoFlush = true;

                //copy training dataset
                trainingIndex = 0;
                for (int c = 0; c < tuples.Length; c++)
                {
                    if ((c >= (testIndex - testTuples.Length)) && (c < testIndex))
                        continue;
                    else if (trainingIndex >= trainingTuples.Length)
                        break;
                    else
                    {
                        trainingTuples[trainingIndex++] = tuples[c];
                        trainingDataset.Write(tuples[c]);
                        trainingDataset.Write("\n");
                    }

                }//for (int c = 0; c < tuples.Length; c++)

                trainingDataset.Flush();
                trainingDataset.Close();

                //delete all files in temp folder
                
                foreach(String tempFile in Directory.GetFiles("output\\temp"))
                   // File.Delete(tempFile);
                
                //generalize training tuples
                trainingFlag = true;
                testFlag = false;
                GeneralizeTrainingTuples();

                //generalize test tuples
                trainingFlag = false;
                testFlag = true;
                GeneralizeTestTuples();

                suppressClasses = class_count_tuples.Count;

                J48_foldAccuracyValues[a] = new ArrayList();
                NaiveBayes_foldAccuracyValues[a] = new ArrayList();
                Logistics_foldAccuracyValues[a] = new ArrayList();
                SVM_foldAccuracyValues[a] = new ArrayList();
                
                for (int c = -1; c < suppressClasses; c++)
                {
                    ////////////////// only to get accuracy of few valeu of k
                    if (c > 2 && c < (suppressClasses - 10))
                        c = c + kInterval;
                    /////////////////

                    if (classifierFlag[J48] == "1")
                        J48_foldAccuracyValues[a].Add(WekaTrainTestEvaluationJ48(c));

                    if (classifierFlag[NAIVE_BAYES] == "1")
                        NaiveBayes_foldAccuracyValues[a].Add(WekaTrainTestEvaluationNaiveBayes(c));

                    if (classifierFlag[LOGISTICS] == "1")
                        Logistics_foldAccuracyValues[a].Add(WekaTrainTestEvaluationLogistics(c));

                    if(classifierFlag[SVM] == "1")
                        SVM_foldAccuracyValues[a].Add(WekaTrainTestEvaluationLibSVM(c));

                    
                }//for (int c = -1; c < suppressClasses; c++)
                
            }//for (int a = 0; a < folds; a++)

            ArrayList temp_class;
            DistinctTuple temp_tuple;

            J48_accuracy = NaiveBayes_accuracy = Logistics_accuracy = LibSVM_accuracy = 0;
            
            
            // minSuppressClasses ++;    //to accommodate origional + generalized dataset
            minSuppressClasses = 0;
            //get number of classes suppressed
            for(int e = 0; e < folds; e++)
            {
                if (classifierFlag[J48] == "1")
                {
                    if (minSuppressClasses == 0)
                        minSuppressClasses = J48_foldAccuracyValues[0].Count;
                    else
                        if (minSuppressClasses > J48_foldAccuracyValues[e].Count)
                            minSuppressClasses = J48_foldAccuracyValues[e].Count;
                }
                else if (classifierFlag[NAIVE_BAYES] == "1")
                {
                    if (minSuppressClasses == 0)
                        minSuppressClasses = NaiveBayes_foldAccuracyValues[0].Count;
                    else
                        if (minSuppressClasses > NaiveBayes_foldAccuracyValues[e].Count)
                            minSuppressClasses = NaiveBayes_foldAccuracyValues[e].Count;
                }
                else if (classifierFlag[LOGISTICS] == "1")
                {
                    if (minSuppressClasses == 0)
                        minSuppressClasses = Logistics_foldAccuracyValues[0].Count;
                    else
                        if (minSuppressClasses > Logistics_foldAccuracyValues[e].Count)
                            minSuppressClasses = Logistics_foldAccuracyValues[e].Count;
                }
                else if (classifierFlag[SVM] == "1")
                {
                    if (minSuppressClasses == 0)
                        minSuppressClasses = SVM_foldAccuracyValues[0].Count;
                    else
                        if (minSuppressClasses > SVM_foldAccuracyValues[e].Count)
                            minSuppressClasses = SVM_foldAccuracyValues[e].Count;
                }
            }//for(int e = 0; e < folds; e++)

            int suppressCount = 0;
            for (int d = 0; d < minSuppressClasses; d++)
            {
               J48_accuracy = NaiveBayes_accuracy = Logistics_accuracy = LibSVM_accuracy = 0;

                for (int c = 0; c < folds; c++)
                {
                    if(classifierFlag[J48] == "1")
                        J48_accuracy += (double) J48_foldAccuracyValues[c][d];
                    if (classifierFlag[NAIVE_BAYES] == "1")
                        NaiveBayes_accuracy += (double) NaiveBayes_foldAccuracyValues[c][d];
                    if (classifierFlag[LOGISTICS] == "1")
                    Logistics_accuracy += (double)Logistics_foldAccuracyValues[c][d];
                    if (classifierFlag[SVM] == "1")
                    LibSVM_accuracy += (double)SVM_foldAccuracyValues[c][d];

                }//for (int c = 0; c < folds; c++)

                J48_accuracy = J48_accuracy / folds;
                NaiveBayes_accuracy = NaiveBayes_accuracy / folds;
                Logistics_accuracy = Logistics_accuracy / folds;
                LibSVM_accuracy = LibSVM_accuracy / folds;

                ////////////////// only to get accuracy of few valeu of k
                if (suppressCount > 2 && suppressCount < (suppressClasses - 10))
                {
                    //temp_class = (ArrayList)class_count_tuples[suppressCount + 6];
                    temp_class = (ArrayList)class_count_tuples[suppressCount + kInterval];
                    suppressCount += kInterval;
                }
                else
                {
                    temp_class = (ArrayList)class_count_tuples[suppressCount + 1];
                    suppressCount++;
                }

                /////////////////
                //temp_class = (ArrayList)class_count_tuples[d + 1];
                temp_tuple = (DistinctTuple)temp_class[0];


                switch(d)
                {
                    //logValues.Write("K = " + (temp_class.Count * temp_tuple.count) + " J48 Accuracy - Origional Dataset " + J48_accuracy);
                    case 0:
                        logValues.WriteLine("");
                        logValues.Write("K = " + temp_tuple.count + " J48 Accuracy - Origional Dataset " + J48_accuracy);
                        logValues.Write("\tK = " + temp_tuple.count + " NaiveBayes Accuracy - Oritional Dataset " + NaiveBayes_accuracy);
                        logValues.Write("\tK = " + temp_tuple.count + " Logistics Accuracy - Oritional Dataset " + Logistics_accuracy);
                        logValues.Write("\tK = " + temp_tuple.count + " SVM Accuracy - Oritional Dataset " + LibSVM_accuracy);

                    break;

                    case 1:
                        logValues.WriteLine("");
                        logValues.Write("K = " + temp_tuple.count + " J48 Accuracy - Generalized Dataset " + J48_accuracy);
                        logValues.Write("\tK = " + temp_tuple.count + " NaiveBayes Accuracy - Generalized Dataset " + NaiveBayes_accuracy);
                        logValues.Write("\tK = " + temp_tuple.count + " Logistics Accuracy - Generalized Dataset " + Logistics_accuracy);
                        logValues.Write("\tK = " + temp_tuple.count + " SVM Accuracy - Generalized Dataset " + LibSVM_accuracy);

                   break;

                    default:

                        logValues.WriteLine("");
                        logValues.Write("Class = " + (suppressCount + 1) + ", K = " + temp_tuple.count + " J48 Accuracy: " + J48_accuracy);
                        logValues.Write("\tClass = " + (suppressCount + 1) + ", K = " + temp_tuple.count + " NaiveBayes Accuracy: " + NaiveBayes_accuracy);
                        logValues.Write("\tClass = " + (suppressCount + 1) + ", K = " + temp_tuple.count + " Logistics Accuracy: " + Logistics_accuracy);
                        logValues.Write("\tClass = " + (suppressCount + 1) + ", K = " + temp_tuple.count + " SVM Accuracy: " + LibSVM_accuracy);
                    break;
                }
                
            }//for (int d = 0; d < minSuppressClasses; d++)

            
        }//public void CalculateTrainingTestWekaClassifierAccuracyValues()


        public double WekaTrainTestEvaluationLibSVM(int suppressClass)
        {
            weka.core.Instances train;
            weka.core.Instances test;

            if (suppressClass == -1)
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\origional_weka_training.arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\origional_weka_test.arff"));
            }
            else if (suppressClass == 0)
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_training_weka.arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_test_weka.arff"));
            }
            else
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\suppress_class_training_SVM_" + suppressClass + ".arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_test_weka.arff"));
            }

            train.setClassIndex(train.numAttributes() - 1);
            test.setClassIndex(test.numAttributes() - 1);

            weka.classifiers.Classifier cls = new weka.classifiers.functions.LibSVM();

            cls.buildClassifier(train);

            weka.classifiers.Evaluation eval = new weka.classifiers.Evaluation(train);
            eval.evaluateModel(cls, test);

            double correct = (eval.correct() / eval.numInstances()) * 100;

            //MessageBox.Show(eval.toSummaryString());

            return correct;

        }//public double WekaTrainTestEvaluationLibSVM()


        public double WekaTrainTestEvaluationLogistics(int suppressClass)
        {
            weka.core.Instances train;
            weka.core.Instances test;

            if (suppressClass == -1)
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\origional_weka_training.arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\origional_weka_test.arff"));
            }
            else if (suppressClass == 0)
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_training_weka.arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_test_weka.arff"));
            }
            else
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\suppress_class_training_" + suppressClass + ".arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_test_weka.arff"));
            }

            train.setClassIndex(train.numAttributes() - 1);
            test.setClassIndex(test.numAttributes() - 1);

            weka.classifiers.Classifier cls = new weka.classifiers.functions.Logistic();

            cls.buildClassifier(train);

            weka.classifiers.Evaluation eval = new weka.classifiers.Evaluation(train);
            eval.evaluateModel(cls, test);

            double correct = (eval.correct() / eval.numInstances()) * 100;

            //MessageBox.Show(eval.toSummaryString());

            return correct;

        }//public double WekaTrainTestEvaluationLogistics()
        
        public double WekaTrainTestEvaluationNaiveBayes(int suppressClass)
        {
            weka.core.Instances train;
            weka.core.Instances test;

            if (suppressClass == -1)
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\origional_weka_training.arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\origional_weka_test.arff"));
            }
            else if (suppressClass == 0)
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_training_weka.arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_test_weka.arff"));
            }
            else
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\suppress_class_training_" + suppressClass + ".arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_test_weka.arff"));
            }

            train.setClassIndex(train.numAttributes() - 1);
            test.setClassIndex(test.numAttributes() - 1);

            weka.classifiers.Classifier cls = new weka.classifiers.bayes.NaiveBayes();

            cls.buildClassifier(train);

            weka.classifiers.Evaluation eval = new weka.classifiers.Evaluation(train);
            eval.evaluateModel(cls, test);

            double correct = (eval.correct() / eval.numInstances()) * 100;

            //MessageBox.Show(eval.toSummaryString());

            return correct;

        }//public double WekaTrainTestEvaluationNaiveBayes()

        public double WekaTrainTestEvaluationJ48(int suppressClass)
        {
            weka.core.Instances train;
            weka.core.Instances test;

            if (suppressClass == -1)
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\origional_weka_training.arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\origional_weka_test.arff"));
            }
            else if (suppressClass == 0)
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_training_weka.arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_test_weka.arff"));
            }
            else
            {
                train = new weka.core.Instances(new java.io.FileReader("output\\temp\\suppress_class_training_"+suppressClass+".arff"));
                test = new weka.core.Instances(new java.io.FileReader("output\\temp\\generalized_test_weka.arff"));
            }

            train.setClassIndex(train.numAttributes() - 1);
            test.setClassIndex(test.numAttributes() - 1);

            weka.classifiers.Classifier cls = new weka.classifiers.trees.J48();

            cls.buildClassifier(train);

            weka.classifiers.Evaluation eval = new weka.classifiers.Evaluation(train);
            eval.evaluateModel(cls, test);

            double correct = (eval.correct() / eval.numInstances()) * 100;

            
            //MessageBox.Show(eval.toSummaryString());

            return correct;

        }//public double WekaTrainTestEvaluation()

        public void GeneralizeTestTuples()
        {
            for (int a = 0; a < totalQIs; a++)
            {
                QIs_Levels[a] = new ArrayList(QIs[a].levels);
                CreateLevels(ref QIs_Levels[a], a);
            }

            //
            //write origional data into weka file format
            //
            CreateOrigionalWekaDataSet();

            //
            //generalize and write generalized data into weka file format
            //
            GeneralizeDataToToMaximumTradeoffLevel(QIs_Levels);

        }//public void GeneralizeTestTuples()

        public void GeneralizeTrainingTuples()
        {

            for (int a = 0; a < totalQIs; a++)
            {
                QIs_Levels[a] = new ArrayList(QIs[a].levels);
                CreateLevels(ref QIs_Levels[a], a);
            }

            //
            //write origional data into weka file format
            //
            CreateOrigionalWekaDataSet();


            //
            //generalize and write generalized data into weka file format
            //
            GeneralizeDataToToMaximumTradeoffLevel(QIs_Levels);

            //count distict tuples after generalization
            CountDistinctTuples(ref generalizedTuples, ref generalizedDistinctTuples);


            //sort tuples first number of distinct tuples and secondly total number of distinct tuples
            SortGeneralizedTuples();

            logValues.WriteLine("");
            logValues.WriteLine("Number of Distinct Tuples (Training Dataset): " + generalizedDistinctTuples.Count.ToString());
            logValues.WriteLine("");

            logValues.WriteLine("Size Equivalence Classes (Training Dataset): " + class_count_tuples.Count.ToString());

            logValues.WriteLine("\nNumber of Tuples in each class (Training Dataset)");

            ArrayList temp_class;
            DistinctTuple temp_tuple;
            for (int a = 1; a <= class_count_tuples.Count; a++)
            {
                temp_class = (ArrayList)class_count_tuples[a];
                temp_tuple = (DistinctTuple)temp_class[0];

                logValues.WriteLine("Class: " + a + " Tuples: " + (temp_class.Count * temp_tuple.count));
            }

            
            int suppressClasses = class_count_tuples.Count;
            
            if (suppressClasses > 0)
                suppressClasses = SuppressClasses(suppressClasses);

            logValues.WriteLine("");


        }//public void GeneralizeTrainingTuples
        
        //Common Methods
        public void CountDistinctTuples(ref ArrayList generalizedTuples, ref ArrayList generalizedDistinctTuples)
        {
            //sensitiveFlag
            String line = "", token = "", generalizedQIs = "", nonGeneralizedIDs = "";
            int firstIndex = 0, lastIndex = 0;
            DistinctTuple distinctTuple;
            bool distinctFound = false;
            String generalizedTuple = "";

            generalizedDistinctTuples = new ArrayList();
            for (int a = 0; a < generalizedTuples.Count; a++)
            {
                line = (String) generalizedTuples[a];

                generalizedQIs = ""; nonGeneralizedIDs = ""; generalizedTuple = "";
                for (int b = 0; b < totalQIs; b++)
                {
                    firstIndex = lastIndex = 0;

                    if (b == 0)
                    {
                        lastIndex = line.IndexOf(',', 0);
                        token = line.Substring(firstIndex, lastIndex);
                    }
                    else
                    {
                        firstIndex = line.IndexOf(',');

                        for (int c = 1; c < b; c++)
                        {
                            firstIndex++;
                            firstIndex = line.IndexOf(',', firstIndex);
                        }

                        firstIndex++;
                        lastIndex = line.IndexOf(',', firstIndex);

                        if (lastIndex > 0)
                            token = line.Substring(firstIndex, lastIndex - firstIndex);
                        else
                            token = line.Substring(firstIndex);

                    }// else of if (b == 0)

                    if (sensitiveFlag[b] == 0)
                        generalizedQIs += token + ",";
                    //else
                    //    nonGeneralizedIDs += token + ",";

                    generalizedTuple += token + ",";

                }//for (int b = 0; b < totalQIs; b++)

                generalizedQIs = generalizedQIs.Substring(0,generalizedQIs.Length - 1);
                //nonGeneralizedIDs = nonGeneralizedIDs.Substring(0, nonGeneralizedIDs.Length - 1);
                generalizedTuple = generalizedTuple.Substring(0, generalizedTuple.Length - 1);

                if (generalizedDistinctTuples.Count == 0)
                {
                    distinctTuple = new DistinctTuple();
                    distinctTuple.count = 1;
                    //distinctTuple.list_of_Non_Generalized_IDs = new ArrayList();
                    //distinctTuple.list_of_Non_Generalized_IDs.Add(nonGeneralizedIDs);
                    distinctTuple.tuple = new ArrayList();
                    distinctTuple.tuple.Add(generalizedTuple);
                    distinctTuple.QIs = generalizedQIs;
                    generalizedDistinctTuples.Add(distinctTuple);
                    continue;
                }

                distinctFound = false;
                for (int c = 0; c < generalizedDistinctTuples.Count; c++)
                {
                    distinctTuple = (DistinctTuple) generalizedDistinctTuples[c];

                    if (distinctTuple.QIs == generalizedQIs)
                    {
                        distinctTuple.count++;
                        //distinctTuple.list_of_Non_Generalized_IDs.Add(nonGeneralizedIDs);
                        distinctTuple.tuple.Add(generalizedTuple);
                        generalizedDistinctTuples[c] = distinctTuple;
                        distinctFound = true;
                        break;
                    }

                }//for (int c = 0; c < generalizedDistinctTuples.Count; c++)

                if (!distinctFound)
                {
                    distinctTuple = new DistinctTuple();
                    distinctTuple.count = 1;
                    //distinctTuple.list_of_Non_Generalized_IDs = new ArrayList();
                    //distinctTuple.list_of_Non_Generalized_IDs.Add(nonGeneralizedIDs);
                    distinctTuple.tuple = new ArrayList();
                    distinctTuple.tuple.Add(generalizedTuple);
                    distinctTuple.QIs = generalizedQIs;
                    generalizedDistinctTuples.Add(distinctTuple);
                }

            }//for (int a = 0; a < generalizedTuples.Count; a++)

        }//public void CountDistinctTuples(ref ArrayList generalizedTuples, ref ArrayList distinct_tuples_count)

        public void SortGeneralizedTuples()
        {
            DistinctTuple tempTupleObj1, tempTupleObj2;

            distinct_tuple_objects = new DistinctTuple[generalizedDistinctTuples.Count];

            double count = 0;
            
            for (int a = 0; a < generalizedDistinctTuples.Count; a++)
            {
                tempTupleObj1 = (DistinctTuple)generalizedDistinctTuples[a];
                count += tempTupleObj1.count;

                for (int b = 0; b < generalizedDistinctTuples.Count-1-a; b++)
                {
                    tempTupleObj1 = (DistinctTuple)generalizedDistinctTuples[b];
                    tempTupleObj2 = (DistinctTuple)generalizedDistinctTuples[b+1];

                    if (tempTupleObj1.count > tempTupleObj2.count)
                    {
                        generalizedDistinctTuples[b + 1] = generalizedDistinctTuples[b];
                        generalizedDistinctTuples[b] = tempTupleObj2;
                    } 
                    
                }//for (int b = 0; b < a; b++)

            }//for (int tempA = 0; a < distinct_tuples_count.Count; tempA++)

            count = 0;
            

            //sort total number of distinct tuples
            ////
            tempTupleObj1 = (DistinctTuple)generalizedDistinctTuples[generalizedDistinctTuples.Count - 1];

            class_count = new int[tempTupleObj1.count + 1];
            class_count_tuples = new Hashtable();
            class_high_low = new Hashtable();

            int tempClassCount = 0;
            int tuple_count = 0;
            ArrayList temp_tuples;
            
            //count size of distinct tuples 
            for (int a = 0; a < generalizedDistinctTuples.Count; a++)
            {
                tempTupleObj1 = (DistinctTuple)generalizedDistinctTuples[a];
                class_count[tempTupleObj1.count]++;
            }

            int h = 1, equivalence_class = 1;
            for (int g = 1; g < class_count.Length; g++)
            {
                if (class_count[g] == 0)
                    continue;

                logValues.WriteLine("Equivalence class " + equivalence_class + ", Tuples from " + h + "-" + ((class_count[g] + h)-1) + ", frequency (" + class_count[g] + ")" + " repeated " + g + " times, " + "Total Tuples in eqivalence class " + equivalence_class + ": " + ((class_count[g] * g)));
                equivalence_class++;
                h += class_count[g];
          
            }


            ////sort size of distinct tuples
            // NO NEED TO SORT - 15/11/2009
            //for (int a = 1; a < class_count.Length; a++)
            //{
            //    for (int b = a; b < class_count.Length; b++)
            //    {
            //        if (class_count[a] < class_count[b])
            //        {
            //            tempClassCount = class_count[a];
            //            class_count[a] = class_count[b];
            //            class_count[b] = tempClassCount;

            //        }//if (class_count[a] > class_count[b])

            //    }// for (int b = 0; b < a; b++)

            //}//for (int a = 1; a < class_count.Length; a++)


            class_count_non_sorted = (int[])class_count.Clone();

            tempClassCount = 0;
            for (int c = 0; c < class_count.Length; c++)
            {
                if (class_count[c] == 0)
                    continue;

                tempClassCount++;
                temp_tuples = new ArrayList();
                               
                for (int d = 0; d < class_count[c]; d++)
                    temp_tuples.Add(generalizedDistinctTuples[tuple_count + d]);
                

                tuple_count += class_count[c];

                class_count_tuples.Add(tempClassCount, temp_tuples);

                //tempTupleObj1 = (DistinctTuple)temp_tuples[0];

            }//for (int c = 0; c < class_count.Length; c++)


            ArrayList tempClass1, tempClass2;
            DistinctTuple tempTuple1, tempTuple2;


            for (int e = 1; e <= class_count_tuples.Count; e++)
            {
                tempClass1 = (ArrayList)class_count_tuples[e];
                tempTuple1 = (DistinctTuple)tempClass1[0];

                for (int f = e; f <= class_count_tuples.Count; f++)
                {

                    tempClass2 = (ArrayList)class_count_tuples[f];
                    tempTuple2 = (DistinctTuple)tempClass2[0];


                    //if ((tempTuple1.count * tempClass1.Count) > (tempTuple2.count * tempClass2.Count))
                    if ((tempTuple1.count) > (tempTuple2.count))
                    {
                        class_count_tuples.Remove(e);
                        class_count_tuples.Remove(f);

                        class_count_tuples.Add(e, tempClass2);
                        class_count_tuples.Add(f, tempClass1);

                        tempClass1 = tempClass2;
                        tempTuple1 = tempTuple2;

                    }//if((tempTuple1.count * tempClass1.Count) > (tempTuple2.count * tempClass2.Count))

                }//for (int f = e; f <= class_count_tuples.Count; f++)

            }//for (int e = 1; e <= class_count_tuples.Count; e++)
            

        }//public void SortGeneralizedTuples()
        
        public void GeneralizeDataToToMaximumTradeoffLevel(ArrayList[] QIs_Levels)
        {
            if(testFlag == false)
                maximumTradeoffLevels = new int[totalQIs];

            ArrayList tempLevels;
            Level tempLevel;
            Group tempGroup;

            //2 4 1 2 1 2 1 1 1         generalization levels using our algo
            //1 1 1 1 1 1 1 1 1
            for (int a = 0; a < totalQIs; a++)
            {
                tempLevels = QIs_Levels[a];
                tempLevel = (Level)tempLevels[0];

                if(testFlag == false)
                    maximumTradeoffLevels[a] = tempLevel.maxTradeoffLevel;

            }//for (int a = 0; a < totalQIsl; a++)

            logValues.WriteLine("");

            DateTime dateTimeStartGeneralization;
            DateTime dateTimeEndGeneralization;

            long millisecondsStartGeneralization;
            long millisecondsEndGeneralization;

            currentProcess = Process.GetCurrentProcess();
            millisecondsStartGeneralization = (long)currentProcess.TotalProcessorTime.TotalMilliseconds;

            dateTimeStartGeneralization = DateTime.Now;

            logValues.WriteLine("generalization of dataset starts at " +
                                                            dateTimeStartGeneralization.Day.ToString() + "/" +
                                                            dateTimeStartGeneralization.Month.ToString() + "/" +
                                                            dateTimeStartGeneralization.Year.ToString() + "/ " +
                                                            dateTimeStartGeneralization.Hour.ToString() + ":" +
                                                            dateTimeStartGeneralization.Minute.ToString() + ":" +
                                                            dateTimeStartGeneralization.Second.ToString() + ":" +
                                                            dateTimeStartGeneralization.Millisecond.ToString());


            String line = "", generalizedLine = "", token = "", classifierToken = "", classifierName = "";
            int firstIndex = 0, lastIndex = 0;
            bool groupFlag = false, classifierFlag = false;
            Int64 classifierCount = 1;


            generalized_QIs_Levels = new Level[totalQIs];
            generalizedTuples = new ArrayList();

            for (int d = 0; d < generalized_QIs_Levels.Length; d++)
                generalized_QIs_Levels[d].levelGroups = new ArrayList();


            int loopCount = 0;

            if (trainingFlag == true)
                loopCount = trainingTuples.Length;
            else if (testFlag == true)
                loopCount = testTuples.Length;
            else
                loopCount = tuples.Length;

            for (int a = 0; a < loopCount; a++)
            {
                if (trainingFlag == true)
                    line = trainingTuples[a];
                else if (testFlag == true)
                    line = testTuples[a];
                else
                    line = tuples[a];

                if (line == null)
                    break;

                for (int b = 0; b < totalQIs; b++)
                {
                    firstIndex = lastIndex = 0;
                    
                    if (b == 0)
                    {
                        lastIndex = line.IndexOf(',',0);
                        token = line.Substring(firstIndex, lastIndex);
                    }
                    else
                    {
                        firstIndex = line.IndexOf(',');

                        for (int c = 1; c < b; c++)
                        {
                            firstIndex++;
                            firstIndex = line.IndexOf(',', firstIndex);
                        }

                        firstIndex++;
                        lastIndex = line.IndexOf(',', firstIndex);

                        if(lastIndex > 0)
                            token = line.Substring(firstIndex, lastIndex - firstIndex);
                        else
                            token = line.Substring(firstIndex);

                    }// else of if (b == 0)

                    //
                    classifierToken = line.Substring(line.LastIndexOf(",")+1);

                    if (sensitiveFlag[b] == 0)
                    {
                        token = GeneralizedAttribute(maximumTradeoffLevels[b], token, QIs_Levels[b]);
                        generalizedLine += token;

                        if (generalized_QIs_Levels[b].levelGroups.Count == 0)
                        {
                            tempLevel = new Level();
                            tempGroup = new Group();
                            
                            tempGroup.classifierNames = new ArrayList();
                            tempGroup.classifierCount = new ArrayList();
                            classifierCount = 1;

                            tempGroup.groupName = token;
                            tempGroup.groupCount = 1;

                            tempGroup.classifierNames.Add(classifierToken);
                            tempGroup.classifierCount.Add(classifierCount);

                            tempGroup.groupCount = 1;

                            tempLevel.levelGroups = new ArrayList();
                            tempLevel.levelGroups.Add(tempGroup);

                            generalized_QIs_Levels[b] = tempLevel;

                        }//if (generalized_QIs_Levels[b].Count == 0)
                        else
                        {
                            tempLevel = (Level) generalized_QIs_Levels[b];

                            groupFlag = false;
                            for (int c = 0; c < tempLevel.levelGroups.Count; c++)
                            {
                                tempGroup = (Group)tempLevel.levelGroups[c];

                                if (tempGroup.groupName == token)
                                {
                                    classifierFlag = false;

                                    for (int d = 0; d < tempGroup.classifierNames.Count; d++)
                                    {
                                        classifierName = (String) tempGroup.classifierNames[d];

                                        if (classifierName == classifierToken)
                                        {
                                            classifierCount = (Int64) tempGroup.classifierCount[d];
                                            classifierCount++;
                                            tempGroup.classifierCount[d] = classifierCount;
                                            tempGroup.groupCount++;

                                            classifierFlag = true;

                                            break;
                                        }
                                        
                                    }//for (int a = 0; a < tempGroup.classifierNames.Count; a++)

                                    if (!classifierFlag)    //classifier not found for generalized value of QI attribute
                                    {
                                        classifierCount = 1;
                                        tempGroup.groupCount++;
                                        tempGroup.classifierNames.Add(classifierToken);
                                        tempGroup.classifierCount.Add(classifierCount);
                                    }
                                    
                                    tempLevel.levelGroups[c] = tempGroup;
                                    generalized_QIs_Levels[b] = tempLevel;

                                    groupFlag = true;

                                    break;
                                    
                                }//if (tempGroup.groupName == token)

                            }//for (int c = 0; c < generalized_QIs_Levels[b].Count; c++)

                            if (!groupFlag)// generalized QI attribute not found
                            {
                                tempGroup = new Group();
                                tempGroup.classifierNames = new ArrayList();
                                tempGroup.classifierCount = new ArrayList();

                                tempGroup.groupName = token;
                                tempGroup.groupCount = 1;

                                classifierCount = 1;
                                tempGroup.classifierNames.Add(classifierToken);
                                tempGroup.classifierCount.Add(classifierCount);

                                tempGroup.groupCount = 1;

                                tempLevel.levelGroups.Add(tempGroup);
                                generalized_QIs_Levels[b] = tempLevel;

                            }//if (!groupFlag)

                        }//else of if (generalized_QIs_Levels[b].Count == 0)

                    }//if (sensitiveFlag[b] == 0)
                    else
                        generalizedLine += token;

                    if (b + 1 < totalQIs)
                        generalizedLine += ",";

                }//for (int b = 0; b < totalQIs; b++)

                generalizedTuples.Add(generalizedLine);
                generalizedLine = "";

            }//for (int a = 0; a < tuples.Length; a++)

            dateTimeEndGeneralization = DateTime.Now;

            logValues.WriteLine("generalization of dataset ends at " +
                                                            dateTimeEndGeneralization.Day.ToString() + "/" +
                                                            dateTimeEndGeneralization.Month.ToString() + "/" +
                                                            dateTimeEndGeneralization.Year.ToString() + "/ " +
                                                            dateTimeEndGeneralization.Hour.ToString() + ":" +
                                                            dateTimeEndGeneralization.Minute.ToString() + ":" +
                                                            dateTimeEndGeneralization.Second.ToString() + ":" +
                                                            dateTimeEndGeneralization.Millisecond.ToString());
            logValues.WriteLine("");

            logValues.WriteLine("Total time taken in generalization " + (dateTimeEndGeneralization - dateTimeStartGeneralization).ToString());

            currentProcess = Process.GetCurrentProcess();
            millisecondsEndGeneralization = (long)currentProcess.TotalProcessorTime.TotalMilliseconds;
            logValues.WriteLine("Total CPU time taken in generalization " + (millisecondsEndGeneralization - millisecondsStartGeneralization).ToString());

            logValues.WriteLine("");


            //////////////////////////////////////////
            //write generalized dataset

            StreamWriter generalized_Data_Weka = null, generalized_Data_Weka_SVM = null;


            if (trainingFlag == true)
                generalized_Data_Weka = new StreamWriter("output\\temp\\generalized_training_weka.arff");
            else if (testFlag == true)
                generalized_Data_Weka = new StreamWriter("output\\temp\\generalized_test_weka.arff");
            else
                generalized_Data_Weka = new StreamWriter("output\\generalized_weka.arff");
            
            generalized_Data_Weka.AutoFlush = true;

            generalized_Data_Weka.WriteLine("%\t Generalized Data File Created at " + DateTime.Now.ToString());

            generalized_Data_Weka.WriteLine("");
            generalized_Data_Weka.WriteLine("");
            generalized_Data_Weka.WriteLine("");

            generalized_Data_Weka.WriteLine("@relation Generalized_dataset_count_" + generalizedTuples.Count);


            for (int b = 0; b < totalQIs; b++)
            {
                //sids
                generalized_Data_Weka.Write("@attribute " + qids[b].InnerText + " {");

                WriteGeneralizationTree(ref generalized_Data_Weka, ref generalized_Data_Weka_SVM, maximumTradeoffLevels[b], QIs_Levels[b]);
                
                generalized_Data_Weka.BaseStream.Position = generalized_Data_Weka.BaseStream.Position - 1;
                generalized_Data_Weka.WriteLine("}");

            }//for (int b = 0; a < QI_GeneralizationTree.Length; a++)

            generalized_Data_Weka.WriteLine("");

            generalized_Data_Weka.WriteLine("@data");


            for (int d = 0; d < generalizedTuples.Count; d++)
            {
                if (generalized_Data_Weka_SVM != null)
                    generalized_Data_Weka_SVM.WriteLine(((String)generalizedTuples[d]).Replace("?","Any"));

                generalized_Data_Weka.WriteLine(generalizedTuples[d]);
            }

            if (generalized_Data_Weka_SVM != null)
                generalized_Data_Weka_SVM.Close();
            
            generalized_Data_Weka.Close();

            ///////////////////////////////////////////

        }//public void GeneralizeDataToToMaximumTradeoffLevel(ArrayList [] QIs_Levels)

        public void WriteGeneralizationTree(ref StreamWriter generalized_Data_Weka, ref StreamWriter generalized_Data_Weka_SVM, int maximumTradeoffLevel, ArrayList QI_Levels)
        {
            Level tempLevel = (Level)QI_Levels[maximumTradeoffLevel];
            Group tempGroup;

            if (generalized_Data_Weka_SVM != null)
                generalized_Data_Weka_SVM.Write("Any" + ",");

            generalized_Data_Weka.Write("Any" + ",");

            for (int a = 0; a < tempLevel.levelGroups.Count; a++)
            {
                tempGroup = (Group)tempLevel.levelGroups[a];

                if (generalized_Data_Weka_SVM != null)
                    generalized_Data_Weka_SVM.Write(tempGroup.groupName + ",");

                generalized_Data_Weka.Write(tempGroup.groupName + ",");

            }//for (int a = 0; a < tempLevel.levelGroups.Count; a++)
        
        }//public void WriteGeneralizationTree(ref StreamWriter generalized_Data_Weka, int maximumTradeoffLevels, ArrayList QI_Levels)
        
        public String GeneralizedAttribute(int maximumTradeoffLevel, String token, ArrayList QI_Level)
        {
            Level tempLevel = (Level)QI_Level[maximumTradeoffLevel];
            Group tempGroup;
            int foundLevel = 0;

            if (maximumTradeoffLevel == (QI_Level.Count - 1))
                return token;

            for (int a = 0; a < tempLevel.levelGroups.Count; a++)
            {
                tempGroup = (Group)tempLevel.levelGroups[a];

                foundLevel = 0;
                if((tempGroup.groupTree.Text == token)||(FoundInGroupTree(tempGroup.groupTree.Nodes, token, ref foundLevel)))
                    return tempGroup.groupTree.Text;

            }//for (int a = 0; a < tempLevel.levelGroups.Count; a++)


            return "";

        }//public String GeneralizedAttribute(int maximumTradeoffLevel, String token, ArrayList QI_Level)
        
        public void InitializeDataFileStream(bool flag)
        {
                String tuple = "";
                int tuplesRead = int.Parse(inputLines[0].InnerText);
                int tuplesCount = 0;

                dataFileStreamReader = new StreamReader(inputFile[0].InnerText);//connect open data file stream 
                tuples = new String[tuplesRead];

                while (dataFileStreamReader.Peek() >= 0)
                {
                    tuple = dataFileStreamReader.ReadLine();     //read complete line from data file
                    
                    tuples[tuplesCount] = tuple;
                    tuplesCount++;
                    
                    if(tuplesCount == tuplesRead)
                        break;
                }

                if (tuplesCount < tuplesRead)
                    MessageBox.Show("Tuples Read "+ tuplesCount+ " instead of "+ tuplesRead);

        }//public void InitializeDataFileStream(void)

        public void InitializeNamesFileStream()
        {
            if (namesFileStreamReader != null)
                namesFileStreamReader.Close();

            //MessageBox.Show(Path.GetDirectoryName(openFileDialogData.FileName));
            //namesFileStream = Path.GetDirectoryName(openFileDialogData.SafeFileName);

            //MessageBox.Show(namesFile[0].InnerText);

            namesFileStreamReader = new StreamReader(namesFile[0].InnerText);

            String line = "", qi = "",generalizationType = "", generalizationTree = "";
            int index1 = 0, index2 = 0, temp1 = 0, temp2 = 0, level = 0, QI_Count = 0, nodeCount = 0;
            TreeNodeCollection QI_ParentNode;

            //QI_GeneralizationTree.Initialize();
            
            while (namesFileStreamReader.Peek() >= 0)
            {
                line = namesFileStreamReader.ReadLine();

                index1 = 0;
                index2 = line.IndexOf("\t")-1;
                index2++;
                qi = line.Substring(index1, index2);
                
                index2++;
                index1 = index2;
                index2 = line.IndexOf("\t",index1)-1;
                index2++;
                generalizationType = line.Substring(index1,index2-index1);

                index2++;
                index1 = index2;
                index2 = line.Length;
                index2--;
                generalizationTree = line.Substring(index1, index2 - index1);
                
                temp1 = temp2 = level = 0;
                
                QI_GeneralizationTree[QI_Count] = new TreeView();
                QI_GeneralizationTree[QI_Count].Nodes.Add(qi);
                
                nodeCount = QI_GeneralizationTree[QI_Count].Nodes.Count;
                QI_ParentNode = QI_GeneralizationTree[QI_Count].Nodes[--nodeCount].Nodes;

                QIs[QI_Count].levels = 0;
                EnumerateGeneralizationTree(generalizationTree, ref temp1, ref temp2, ref level, ref QI_ParentNode, ref QIs[QI_Count]);
                QIs[QI_Count].levels++;

                QI_Count++;

                if (QI_Count == totalQIs)
                    break;

            }//while (namesFileStreamReader.Peek() != null)


            //level = -2;// -1 because the name of QI is on top level in generalization tree
            //for (int a = 0; a < QI_GeneralizationTree.Length; a++)
            //{
            //    nodeCount = QI_GeneralizationTree[a].Nodes.Count;
                
            //    level = -2;
            //    for (int b = 0; b < nodeCount; b++)
            //        ReadWholeGeneralizationTree(QI_GeneralizationTree[a].Nodes, ref level);

            //}//for (int a = 0; a < QI_GeneralizationTree.Length; a++)
            

        }//public void InitializeNamesFileStream()

        public int EnumerateGeneralizationTree(String generalizationTree, ref int index1, ref int index2, ref int level, ref TreeNodeCollection QI_ParentNode, ref QI qi)
        {

            String nextNode = "", tempNextNode = "", parentNode = "";
            char nextCharacter = ' ';
            bool open = false;
            int returnLevel = 0, tempIndex1 = 0, tempIndex2 = 0, nodeCount = 0, nodeIndex = 0;
            TreeNodeCollection QI_NextNode = null;

            while (true)
            {

                if (index2 < 0) //no more parent node
                    break;

                tempIndex1 = index1;
                index1 = generalizationTree.IndexOf("{", index2);
                tempIndex2 = index2;
                index2 = generalizationTree.IndexOf("{", index1 + 1);

                if(index2 > 0)
                    nextNode = generalizationTree.Substring(index1 + 1, (index2 - index1) - 1);
                else
                    nextNode = generalizationTree.Substring(index1 + 1);

                if (nextNode.IndexOf("}") < 0)  //true for parent nodes
                {
         //           MessageBox.Show("Level " + level + " Node Value: " + nextNode);

                    if (returnLevel >= 0)   //true when child node exists for parent node
                    {
                        tempNextNode = nextNode;
                        tempNextNode = tempNextNode.Replace("}", "");
                        tempNextNode = tempNextNode.Replace(" ", "");

                        QI_ParentNode.Add(tempNextNode);
                        nodeIndex = QI_ParentNode.Count;
                        QI_NextNode = QI_ParentNode[--nodeIndex].Nodes;
                    }
                    else    //true when no more child node exists BUT there exists more nodes on same level
                    {
                        index1 = tempIndex1;
                        index2 = tempIndex2;
                    }

                    level++;
                    parentNode = nextNode;
                    returnLevel = EnumerateGeneralizationTree(generalizationTree, ref index1, ref index2, ref level, ref QI_NextNode, ref qi);

                    if (returnLevel > 0)
                    {
                        level--;
                        return --returnLevel;
                    }
                    
                }//if (nextNode.IndexOf("}")
                else    //true for child nodes
                {   
                    while (true)
                    {
                        //               MessageBox.Show("Level: " + level + " Deepest Node " + nextNode);
                        tempNextNode = nextNode;
                        tempNextNode = tempNextNode.Replace("}", "");
                        tempNextNode = tempNextNode.Replace(" ", "");

                        QI_ParentNode.Add(tempNextNode);

                        if (qi.levels < level)
                            qi.levels = level;


                        if (index2 < 0)     //all child nodes completed
                        {
                            level--;
                            return 0;
                        }
                        
                        tempIndex1 = index2;
                        open = false;
                        nodeCount = 0;
                        returnLevel = 0;

                        tempIndex2 = nextNode.Length;
                        tempIndex2--;

                        while (true)            //count } in current node to measure that how levels to return
                        {
                            nextCharacter = nextNode[tempIndex2--];

                            if (nextCharacter == '}')
                                returnLevel++;
                            else if (nextCharacter != ' ')
                                break;
                        }

                        
                        while (true)
                        {
                            //get each character from current node and analyze it

                            if (tempIndex1 == generalizationTree.Length)    //true for last child node of the tree
                                nextCharacter = generalizationTree[--tempIndex1];
                            else
                                nextCharacter = generalizationTree[tempIndex1++];

                            if (nextCharacter == '}' && open == true)   //count the child nodes on same level
                            {
                                open = false;
                                nodeCount++;

                                if (nodeCount == 1 && returnLevel == 1) //true when there exists another node on same level 
                                    break;
                            }
                            else if (nextCharacter == '{' && open == false)
                                open = true;
                            else if (nextCharacter == '{' && open == true)  //true when there is child node on this level
                            {
                                if (returnLevel >= 1)    //if returnLevel == 0 than no decrease in level because when function returns a SINGLE level is automatically decreased
                                    level--;

                                returnLevel -= 2;       //determines how many levels further return after this return, one } denotes to this level and next } denotes to the level where function will return, that's why -2 because these two levels are automatically returned

                                return returnLevel;
                              
                            }
                            else if (nextCharacter == '}' && open == false) //true when this levels ends without having any other child node
                            {

                                if (returnLevel >= 1)
                                    level--;

                              
                                returnLevel -= 2;

                                return returnLevel;
                              
                            }//else if (nextCharacter == '}' && open == false)

                        }// while (true)
                        
                        index1 = generalizationTree.IndexOf("{", index2);
                        index2 = generalizationTree.IndexOf("{", index1 + 1);
                       
                        if (index2 > 0)
                            nextNode = generalizationTree.Substring(index1 + 1, (index2 - index1) - 1);
                        else
                            nextNode = generalizationTree.Substring(index1 + 1);

                    }//while(true)

                }//end else

            }//while(true)

            return 0;

        }//public void EnumerateReadGeneralizationTree(ref String generalizationTree)
        
        public void ReadWholeGeneralizationTree(TreeNodeCollection QI_Parent, ref int level)
        {
            int nodeCount = QI_Parent.Count;

            if (nodeCount == 0)
                return;

            level++;
            for (int a = 0; a < nodeCount; a++)
            {
                MessageBox.Show("Level: " + level + " Value: " + QI_Parent[a].Text);
                ReadWholeGeneralizationTree(QI_Parent[a].Nodes, ref level);
            }
            level--;

        }//void ReadWholeGeneralizationTree(TreeNodeCollection QI_Parent)

        public Group InitializeNumericGroup(String groupGroup, int minValue, int maxValue)
        {
            Group tempGroup = new Group();

            tempGroup.groupName = groupGroup;
            tempGroup.groupMinValue = minValue;
            tempGroup.groupMaxValue = maxValue;

            return tempGroup;

        }//public Group InitializeNumericGroup(String groupGroup, int minValue, int maxValue)

        public Group InitializeTextGroup(String groupName)
        {
            Group tempGroup = new Group();

            tempGroup.groupName = groupName;

            return tempGroup;

        }//public Group InitializeTextGroup(String groupName)

        public double CalculateLevelEntropy(ref ArrayList level)
        {
            Group group;
            double levelEntropy = 0;
            long groupCount = 0;

            for (int a = 0; a < level.Count; a++)
            {
                group = (Group)level[a];
                group.entropy = CalculateGroupEntropy(ref group);

                if (group.entropy < 0 || group.entropy > 1)
                    MessageBox.Show("ERROR: Group = " + group.groupName + " Entropy = " + group.entropy);

                level[a] = group;
                groupCount += group.groupCount;
                levelEntropy += group.entropy;
            }

            //MessageBox.Show("level entropy " + levelEntropy.ToString()+ " group count = " + groupCount.ToString());
            //ACCUMMULATIVE ENTROPY OF LEVEL BECOMES MORE THAN 1???
            return levelEntropy;

        }//public void CalculateLevelEntropy(ArrayList level)

        public double CalculateGroupEntropy(ref Group group)
        {
            double groupEntropy = 0;
            
            //if (group.groupCount == 0)
            //    return groupEntropy;

            if (group.groupCount == 0)
            {
                if (trainingFlag == false && testFlag == false)
                    groupEntropy = (double)0.99 / tuples.Length;
                if (trainingFlag == true && testFlag == false)
                    groupEntropy = (double)0.99 / trainingTuples.Length;
                if (trainingFlag == false && testFlag == true)
                    groupEntropy = (double)0.99 / testTuples.Length;
            }
            else
            {
                if (trainingFlag == false && testFlag == false)
                    groupEntropy = (double)group.groupCount / tuples.Length;
                if (trainingFlag == true && testFlag == false)
                    groupEntropy = (double)group.groupCount / trainingTuples.Length;
                if (trainingFlag == false && testFlag == true)
                    groupEntropy = (double)group.groupCount / testTuples.Length;
            }

            groupEntropy = (double)groupEntropy * Math.Log(groupEntropy, 2);

            group.entropy = -groupEntropy;

            return -groupEntropy;

        }//public double CalculateGroupEntropy(Group group)

        public double GetMaximumEntropy(Hashtable levels_Tree)
        {
            double maxEntropy = 0;
            Level tempLevel;

            for (int a = 1; a <= levels_Tree.Count; a++)
            {
                tempLevel = (Level)levels_Tree[a.ToString()];

                if (tempLevel.levelEntropy > maxEntropy)
                    maxEntropy = tempLevel.levelEntropy;
            }

            return maxEntropy;

        }//public double GetMaximumEntropy(Hashtable levels_Tree)

        public double CalculateLevelSpecLoss(double levelEntropy, double maxEntropy)
        {

            return (double)(1 / maxEntropy) * (maxEntropy - levelEntropy);

        }

        public double CalculateTargetAttributeEntropy(ArrayList level1) //0-100
        {
            double targetAttributeEntropy = 0;
            
            Group tempGroup = (Group)level1[0];

            double[] tempClassifierCalculation = new double[tempGroup.classifierNames.Count];
            Int64 classifierCount = 0;
            
            //if (tempGroup.groupCount == 0)
            //    return targetAttributeEntropy;

            for (int a = 0; a < tempGroup.classifierNames.Count; a++)
            {
                classifierCount = (Int64)tempGroup.classifierCount[a];

                if (tempGroup.groupCount == 0)
                    tempClassifierCalculation[a] = (double)classifierCount / 0.5;
                else
                    tempClassifierCalculation[a] = (double)classifierCount / tempGroup.groupCount;

                tempClassifierCalculation[a] *= (double)Math.Log(tempClassifierCalculation[a], 2);

            }//for (int a = 0; a < tempGroup.classifierNames.Count; a++)


            for (int b = 0; b < tempClassifierCalculation.Length; b++)
                targetAttributeEntropy += tempClassifierCalculation[b];

            
            return -targetAttributeEntropy;

        }//public double CalculateTargetAttributeEntropy(ArrayList level1) //0-100

        public double CalculateLevelMutualInformation(ArrayList levelGroups, double targetAttributeEntropy)
        {
            Group tempGroup;
            Int64 groupCount = 0, classifierCount = 0; ;
            double levelMutualInformation = 0, tempCalculation = 0;

            for (int a = 0; a < levelGroups.Count; a++)
            {
                tempGroup = (Group)levelGroups[a];

                if (tempGroup.groupCount == 0)
                {
                    //tempCalculation = (double)0.99 / 0.5;
                    //tempCalculation = -1 * (double)Math.Log(tempCalculation, 2);


                    //if (trainingFlag == false && testFlag == false)
                    //    tempCalculation *= (double)0.000001 / tuples.Length;
                    //if (trainingFlag == true && testFlag == false)
                    //    tempCalculation *= (double)0.000001 / trainingTuples.Length;
                    //if (trainingFlag == false && testFlag == true)
                    //    tempCalculation *= (double)0.000001 / testTuples.Length;

                    //levelMutualInformation += tempCalculation;
                    //tempCalculation = 0;

                    continue;
                }

                for (int b = 0; b < tempGroup.classifierNames.Count; b++)
                {
                    classifierCount = (Int64) tempGroup.classifierCount[b];
                    groupCount += classifierCount;

                    tempCalculation = 0;

                    if (classifierCount > 0)
                    {
                        tempCalculation = (double)classifierCount / tempGroup.groupCount;
                        tempCalculation = -1 * (double)Math.Log(tempCalculation, 2);

                        if (trainingFlag == false && testFlag == false)
                            tempCalculation *= (double)classifierCount / tuples.Length;
                        if(trainingFlag == true && testFlag == false)
                            tempCalculation *= (double)classifierCount / trainingTuples.Length;
                        if (trainingFlag == false && testFlag == true)
                            tempCalculation *= (double)classifierCount / testTuples.Length;

                    }//if (classifierCount > 0)
                    else
                    {
                        //tempCalculation = (double)0.000001 / tempGroup.groupCount;
                        //tempCalculation = -1 * (double)Math.Log(tempCalculation, 2);

                        //if (trainingFlag == false && testFlag == false)
                        //    tempCalculation *= (double)0.000001 / tuples.Length;
                        //if (trainingFlag == true && testFlag == false)
                        //    tempCalculation *= (double)0.000001 / trainingTuples.Length;
                        //if (trainingFlag == false && testFlag == true)
                        //    tempCalculation *= (double)0.000001 / testTuples.Length;
                        
                    }// else of if (classifierCount > 0)

                    levelMutualInformation += tempCalculation;
                    
                }//for (int b = 0; b < tempGroup.classifierNames.Count; b++)

            }//for (int a = 0; a < levelGroups.Count; a++)

            levelMutualInformation = targetAttributeEntropy - levelMutualInformation;
            //levelGain = targetAttributeEntropy - levelGain;
            
            return levelMutualInformation;

        }//public double CalculateLevelMutualInformation(ArrayList levelGroups)

        public double CalculatemaxlevelDifferenceMutualInformationNormalized(Hashtable levels_Tree)
        {
            double maxlevelDifferenceMutualInformationNormalized = 0;
            double tempDifference = 0;

            Level tempLevel_Lowest = (Level)levels_Tree[levels_Tree.Count.ToString()];
            Level tempLevel_Current = (Level)levels_Tree["1"];

            maxlevelDifferenceMutualInformationNormalized = tempLevel_Current.levelMutualInformationNormalized - tempLevel_Lowest.levelMutualInformationNormalized;

            for (int a = 2; a <= levels_Tree.Count; a++)
            {
                tempLevel_Current = (Level)levels_Tree[a.ToString()];

                tempDifference = tempLevel_Current.levelMutualInformationNormalized - tempLevel_Lowest.levelMutualInformationNormalized;

                if (tempDifference > maxlevelDifferenceMutualInformationNormalized)
                    maxlevelDifferenceMutualInformationNormalized = tempDifference;

            }

            return maxlevelDifferenceMutualInformationNormalized;

        }

        public double CalculateClassGain(Level level, double maxMutualInformationNormalized, double maxlevelDifferenceMutualInformationNormalized)
        {
            double classGain = 0;

            classGain = maxlevelDifferenceMutualInformationNormalized != 0 ? (double)(level.levelMutualInformationNormalized - maxMutualInformationNormalized) / maxlevelDifferenceMutualInformationNormalized : -0;
            //classGain = (double)(level.levelMutualInformationNormalized - maxMutualInformationNormalized) / maxlevelDifferenceMutualInformationNormalized;

            return classGain;

        }

        public int CalculateMaximumTradeoffLevel(ref ArrayList levels, ref int []sensitiveFlag)
        {
            int maximumTradeoffLevel = 1;
            double tempLevelTradeoff = 0;
            Level tempLevel;




            tempLevel = (Level)levels[0];
            tempLevelTradeoff = tempLevel.levelTradeoff;

            for (int a = 1; a < levels.Count; a++)
            {
                tempLevel = (Level)levels[a];

                if (tempLevelTradeoff <= tempLevel.levelTradeoff)
                {
                    tempLevelTradeoff = tempLevel.levelTradeoff;
                    maximumTradeoffLevel = a;
                }
            }

            for (int b = 0; b < levels.Count; b++)
            {
                tempLevel = (Level)levels[b];
                tempLevel.maxTradeoffLevel = maximumTradeoffLevel;

                levels[b] = tempLevel;
            }

            return maximumTradeoffLevel;

        }//public int CalculateAgeMaximumTradeoffLevel(Hashtable age_Levels_Tree)

        public double CalculateMaxNormalizedMutualInformation(Hashtable levels_Tree)
        {
            double maxNormalizedMutualInformation = 0;

            Level tempLevel;

            for (int a = 1; a < levels_Tree.Count; a++)
            {
                tempLevel = (Level)levels_Tree[a.ToString()];

                if (tempLevel.levelMutualInformationNormalized > maxNormalizedMutualInformation)
                    maxNormalizedMutualInformation = tempLevel.levelMutualInformationNormalized;
            }

            return maxNormalizedMutualInformation;

        }

        public void CountHighLow(String line, ref Group group)
        {

            Int64 existingClassifierCount = 1;
            int lastIndex = line.LastIndexOf(',');          //get index of last occurance of ','
            //lastIndex += 2;                           //un-comment if want to use adult_32561.data    to discard ',' and ' '
            lastIndex++;
            line = line.Substring(lastIndex, line.Length - lastIndex);

            if (group.classifierNames == null || group.classifierCount == null)
            {
                group.classifierCount = new ArrayList();
                group.classifierNames = new ArrayList();

                group.classifierNames.Add(line);
                group.classifierCount.Add(existingClassifierCount);
                return;
            }

            String existingClassifier = "";
            bool classifierFound = false;

            for (int a = 0; a < group.classifierNames.Count; a++)
            {
                existingClassifier = (String) group.classifierNames[a];

                if (existingClassifier == line)
                {
                    existingClassifierCount = (Int64)group.classifierCount[a];
                    existingClassifierCount++;
                    group.classifierCount[a] = existingClassifierCount;
                    classifierFound = true;
                    break;
                }

            }//for (int a = 0; a < group.classifierNames.Count; a++)

            if (!classifierFound)
            {
                group.classifierNames.Add(line);
                group.classifierCount.Add(existingClassifierCount);
            }

            return;

        }//public void CountHighLow(String line, ref Group group)

        public void InitializeGeneralizedDataFileStream(bool open_flag, bool append_flag)
        {
            if (open_flag)
                generalizedDataFileStreamWriter = new StreamWriter("generalized_adult_45222.data", append_flag);
            else
                generalizedDataFileStreamWriter.Close();

        }//public void InitializeGeneralizedDataFileStream()

        public int SuppressClasses(int numClasses)
        {
            ArrayList temp_class = new ArrayList();
            ArrayList temp_suppressed = new ArrayList();
            SuppressedClass tempSuppressedClass;
            bool suppressionBreakFlag = false;

            suppressedClasses_QIs = new ArrayList[totalQIs];


            for (int d = 0; d < suppressedClasses_QIs.Length; d++)
                suppressedClasses_QIs[d] = new ArrayList(new SuppressedClass[numClasses + 1]);

                       
            DistinctTuple temp_tuple = new DistinctTuple();

            class_count_tuples_suppressed = new Hashtable();

            int firstIndex = 0;
            int lastIndex = 0;

            String line = "", classifierToken = "", token = "", suppressedTuple = "", classifierName = "";
            Int64 classifierCount = 0;

            Group tempGroup, tempGroup_Suppressed;
            Level tempLevel;

            StreamWriter suppress_writer = null, suppress_writer_SVM = null;

            int a = 0;
            int accumulative_count = 0;
            int current_count = 0;
                        

            class_count_tuples_suppressed = (Hashtable)class_count_tuples.Clone();
            generalized_QIs_Suppressed_Levels = new Level[generalized_QIs_Levels.Length];


            for (int h = 0; h < generalized_QIs_Levels.Length; h++)
            {
                generalized_QIs_Suppressed_Levels[h].levelGroups = (ArrayList)generalized_QIs_Levels[h].levelGroups.Clone(); ;

                for (int i = 0; i < generalized_QIs_Levels[h].levelGroups.Count; i++)
                {
                    tempGroup = (Group)generalized_QIs_Levels[h].levelGroups[i];
                    tempGroup_Suppressed = (Group) generalized_QIs_Suppressed_Levels[h].levelGroups[i];

                    tempGroup_Suppressed.classifierNames = (ArrayList)tempGroup.classifierNames.Clone();
                    tempGroup_Suppressed.classifierCount = (ArrayList)tempGroup.classifierCount.Clone();

                    generalized_QIs_Suppressed_Levels[h].levelGroups[i] = tempGroup_Suppressed;

                }//for (int i = 0; i < generalized_QIs_Levels[h].levelGroups.Count; i++)

            }//for (int h = 0; h < generalized_QIs_Levels.Length; h++)


            logValues.WriteLine("");


            DateTime dateTimeSuppressionStart;
            DateTime dateTimeClassSuppressionStart;
            DateTime dateTimeSuppressionEnd;
                        

            long milliSecondsSuppressionStart = 0;
            long milliSecondsClassSuppressionStart = 0;
            long milliSecondsSuppressionEnd = 0;
            long milliSecondsTotalCPUTime = 0;


            currentProcess = Process.GetCurrentProcess();
            milliSecondsSuppressionStart = (long)currentProcess.TotalProcessorTime.TotalMilliseconds;

            dateTimeSuppressionStart = DateTime.Now;

            logValues.WriteLine("Suppression of " + numClasses.ToString() +  " equivalence classes starts at " +
                                                            dateTimeSuppressionStart.Day.ToString() + "/" +
                                                            dateTimeSuppressionStart.Month.ToString() + "/" +
                                                            dateTimeSuppressionStart.Year.ToString() + "/ " +
                                                            dateTimeSuppressionStart.Hour.ToString() + ":" +
                                                            dateTimeSuppressionStart.Minute.ToString() + ":" +
                                                            dateTimeSuppressionStart.Second.ToString() + ":" +
                                                            dateTimeSuppressionStart.Millisecond.ToString());

            logValues.WriteLine("");


            currentProcess = Process.GetCurrentProcess();
            milliSecondsSuppressionStart = (long)currentProcess.TotalProcessorTime.TotalMilliseconds;

            XmlNodeList write_suppress_dataset = xDoc.GetElementsByTagName("writeSuppressDataset");
            String fWriteSuppressDataset = write_suppress_dataset[0].InnerText;

                        
            for (a = 1; a <= numClasses; a++)
            {
                //if (class_count[a] == 0) { continue; }

                //if (a == 35)
                //    a = a; ;
               
                temp_class = (ArrayList)class_count_tuples[a];
                temp_suppressed = new ArrayList();

                currentProcess = Process.GetCurrentProcess();
                milliSecondsClassSuppressionStart = (long)currentProcess.TotalProcessorTime.TotalMilliseconds;

                dateTimeClassSuppressionStart = DateTime.Now;

                //logValues.WriteLine("Suppression of equivalence class " + a.ToString() + " starts at " +
                //                                dateTimetemp.Day.ToString() + "/" +
                //                                dateTimetemp.Month.ToString() + "/" +
                //                                dateTimetemp.Year.ToString() + "/ " +
                //                                dateTimetemp.Hour.ToString() + ":" +
                //                                dateTimetemp.Minute.ToString() + ":" +
                //                                dateTimetemp.Second.ToString() + ":" +
                //                                dateTimetemp.Millisecond.ToString());

               
                //class_count_tuples_suppressed
                for (int b = 0; b < temp_class.Count; b++)
                {
                    temp_tuple = (DistinctTuple)temp_class[b];

                    for (int c = 0; c < temp_tuple.tuple.Count; c++)
                    {

                        line = (String)temp_tuple.tuple[c];

                        for (int d = 0; d < totalQIs; d++)
                        {
                            firstIndex = lastIndex = 0; 
                            if (d == 0)
                            {
                                lastIndex = line.IndexOf(',', 0);
                                token = line.Substring(firstIndex, lastIndex);
                            }
                            else
                            {
                                firstIndex = line.IndexOf(',');

                                for (int e = 1; e < d; e++)
                                {
                                    firstIndex++;
                                    firstIndex = line.IndexOf(',', firstIndex);
                                }

                                firstIndex++;
                                lastIndex = line.IndexOf(',', firstIndex);

                                if (lastIndex > 0)
                                    token = line.Substring(firstIndex, lastIndex - firstIndex);
                                else
                                    token = line.Substring(firstIndex);


                            }// else of if (d == 0)

                            //
                            classifierToken = line.Substring(line.LastIndexOf(",") + 1);


                            if (sensitiveFlag[d] == 1)
                            {
                                suppressedTuple += token + ",";
                                continue;
                            }

                            tempLevel = (Level)generalized_QIs_Suppressed_Levels[d];

                            for (int f = 0; f < tempLevel.levelGroups.Count; f++)
                            {
                                tempGroup = (Group)tempLevel.levelGroups[f];

                                if (tempGroup.groupName == token)
                                {
                                    for (int g = 0; g < tempGroup.classifierNames.Count; g++)
                                    {
                                        classifierName = (String) tempGroup.classifierNames[g];

                                        if (classifierName == classifierToken)
                                        {
                                            classifierCount = (Int64) tempGroup.classifierCount[g];
                                            classifierCount--;
                                            tempGroup.classifierCount[g] = classifierCount;

                                            tempGroup.groupCount--;
                                            tempLevel.levelGroups[f] = tempGroup;

                                            suppressedTuple += "?,";

                                            break;
                                        }

                                    }//for (int g = 0; g < tempGroup.classifierNames.Count; g++)
                                    
                                }//if (tempGroup.groupName == token)

                            }//for (int f = 0; f < tempLevel.levelGroups.Count; f++)

                            generalized_QIs_Suppressed_Levels[d] = tempLevel;

                        }//for (int d = 0; d < totalQIs; d++)

                        suppressedTuple = suppressedTuple.Substring(0, suppressedTuple.Length - 1);
                        temp_tuple.tuple[c] = suppressedTuple;
                        suppressedTuple = "";
                        current_count += temp_tuple.count;

                    }//for (int c = 0; c < temp_tuple.tuple.Count; c++)

                    temp_suppressed.Add(temp_tuple);

                }//for (int b = 0; b < temp_class.Count; b++)

                
                dateTimeSuppressionEnd = DateTime.Now;

                //logValues.WriteLine("Suppression of class " + a.ToString() + " ends at " +
                //                                            dateTimeEnd.Day.ToString() + "/" +
                //                                            dateTimeEnd.Month.ToString() + "/" +
                //                                            dateTimeEnd.Year.ToString() + "/ " +
                //                                            dateTimeEnd.Hour.ToString() + ":" +
                //                                            dateTimeEnd.Minute.ToString() + ":" +
                //                                            dateTimeEnd.Second.ToString() + ":" +
                //                                            dateTimeEnd.Millisecond.ToString());

                logValues.WriteLine("Total time taken in suppressing class " + a.ToString() + " " + (dateTimeSuppressionEnd - dateTimeClassSuppressionStart).ToString());

                currentProcess = Process.GetCurrentProcess();
                milliSecondsSuppressionEnd = (long)currentProcess.TotalProcessorTime.TotalMilliseconds;
                milliSecondsTotalCPUTime += milliSecondsSuppressionEnd - milliSecondsClassSuppressionStart;
                logValues.WriteLine("Total CPU time taken in suppres class " + a.ToString() + " " + (milliSecondsSuppressionEnd - milliSecondsClassSuppressionStart).ToString());

                

                accumulative_count += current_count;
                current_count = 0;

                class_count_tuples_suppressed.Remove(a);
                class_count_tuples_suppressed.Add(a, temp_suppressed);

                //write data into weka file after each class suppression
                if (fWriteSuppressDataset == "1")
                {
                    InitializeSuppressWekaDataSet(ref suppress_writer, ref suppress_writer_SVM, a);
                    WriteSuppressWekaDataSet(ref suppress_writer, ref suppress_writer_SVM, class_count_tuples_suppressed, a, numClasses);
                }

                if (trainingFlag == true)
                {
                    InitializeSuppressWekaDataSet(ref suppress_writer, ref suppress_writer_SVM, a);
                    WriteSuppressWekaDataSet(ref suppress_writer, ref suppress_writer_SVM, class_count_tuples_suppressed, a, numClasses);
                }

                for (int g = 0; g < totalQIs; g++)
                {
                    if (sensitiveFlag[g] == 1)
                        continue;

                    temp_suppressed = (ArrayList)suppressedClasses_QIs[g];
                    tempSuppressedClass = (SuppressedClass)temp_suppressed[a];

                    CalculateDistributionDistance(g , ref tempSuppressedClass, generalized_QIs_Levels[g], generalized_QIs_Suppressed_Levels[g], QIs_Levels[g]);
                    temp_suppressed[a] = tempSuppressedClass;
                    suppressedClasses_QIs[g] = temp_suppressed;

                    if (tempSuppressedClass.distribution_distance_normalized > double.Parse(KullbackLeiblerDivergence[0].InnerText))
                    {
                        logValues.WriteLine("Suppression Loop broken after class " + a.ToString() + " : " + qiNames[g] + " : Kullback-Leibler Divergence " + tempSuppressedClass.distribution_distance_normalized.ToString());
                        suppressionBreakFlag = true;
                        break;
                    }

                    tempSuppressedClass.info_gain_ratio = CalculateInformationGainRatio(a, g, generalized_QIs_Levels[g], generalized_QIs_Suppressed_Levels[g], QIs_Levels[g]);

                    temp_suppressed[a] = tempSuppressedClass;
                    suppressedClasses_QIs[g] = temp_suppressed;

                    if (tempSuppressedClass.info_gain_ratio > Math.Abs(double.Parse(NormalisedMutualInformation[0].InnerText)))
                    {
                        logValues.WriteLine("Suppression Loop broken after class " + a.ToString() + " : " + qiNames[g] + " : Normalised Mutual Information " + tempSuppressedClass.distribution_distance_normalized.ToString());
                        suppressionBreakFlag = true;
                        break;
                    }

                }//for (int g = 0; g < totalQIs; g++)

                if (suppressionBreakFlag)
                    break;
                
            }//for (int a = 1; a <= numClasses; a++)

            logValues.WriteLine("");

            dateTimeSuppressionEnd = DateTime.Now;

            logValues.WriteLine("Suppression of " + a.ToString() + " classes ends at " +
                                                        dateTimeSuppressionEnd.Day.ToString() + "/" +
                                                        dateTimeSuppressionEnd.Month.ToString() + "/" +
                                                        dateTimeSuppressionEnd.Year.ToString() + "/ " +
                                                        dateTimeSuppressionEnd.Hour.ToString() + ":" +
                                                        dateTimeSuppressionEnd.Minute.ToString() + ":" +
                                                        dateTimeSuppressionEnd.Second.ToString() + ":" +
                                                        dateTimeSuppressionEnd.Millisecond.ToString());

            //logValues.WriteLine("Total time taken in suppressing class " + a.ToString() + " " + (dateTimeSuppressionEnd - dateTimeSuppressionStart).ToString());

            //currentProcess = Process.GetCurrentProcess();
            //milliSecondsSuppressionEnd = (long)currentProcess.TotalProcessorTime.TotalMilliseconds;
            logValues.WriteLine("Suppression of " + a.ToString() + " classes took CPU milliseconds " + milliSecondsTotalCPUTime.ToString());

            return a - 1;

        }//public int SuppressClasses(int numClasses)

        //Normalised Mutual Information
        public double CalculateInformationGainRatio(int classNum, int QIindex, Level class_High_Low_Count, Level Suppress_class_High_Low_Count, ArrayList Levels_tree)
        {
            //qiNames[QIindex]
            double infoGainRatio = 0;

            Group temp_group;
            Group temp_group_suppressed;

            Group temp_targetAttribute = new Group();
            temp_targetAttribute.classifierCount = new ArrayList();
            temp_targetAttribute.classifierNames = new ArrayList();

            ArrayList temp_targetAttributeGroup = new ArrayList();
            ArrayList temp_suppressedLevel = new ArrayList();

            Group suppress_count = new Group();
            suppress_count.classifierCount = new ArrayList();
            suppress_count.classifierNames = new ArrayList();

            suppress_count.groupName = "?";

            long record_count = 0;
            Int64 tempClassifierCount = 0, tempClassifierCount_Suppressed = 0, suppressGroupCount = 0, tempCount = 0;
            bool suppressFlag = false;

            //calculate the number of ? by taking difference of "class_High_Low_Count" and "Suppress_class_High_Low_Count"
            //and calculate the corsponding value of "high" and "low" of ? in the same way of taking difference

            if (classNum == 3 || classNum == 16)
            {
                logValues.WriteLine("");
                logValues.WriteLine("=== Intermediate values during Information Gain Ratio ====");
                logValues.WriteLine("QI = " + this.qiNames[QIindex] + ", Class = " + classNum);
            }
            
            for(int a = 0; a < class_High_Low_Count.levelGroups.Count; a++)
            {
                temp_group = (Group)class_High_Low_Count.levelGroups[a];
                temp_group_suppressed = (Group)Suppress_class_High_Low_Count.levelGroups[a];

                if (classNum == 3 || classNum == 16)
                {
                    logValues.WriteLine("QI Group = " + temp_group.groupName + "(without suppression), Count = " + temp_group.groupCount);
                    logValues.WriteLine("QI Group = " + temp_group_suppressed.groupName + "(with suppression), Count = " + temp_group_suppressed.groupCount);
                }

                for (int b = 0; b < temp_group.classifierNames.Count; b++)
                {
                    tempClassifierCount = (Int64)temp_group.classifierCount[b];
                    tempClassifierCount_Suppressed = (Int64)temp_group_suppressed.classifierCount[b];

                    record_count += tempClassifierCount;

                    tempClassifierCount_Suppressed = tempClassifierCount - tempClassifierCount_Suppressed;
                                        
                    suppressFlag = false;
                    for (int c = 0; c < suppress_count.classifierNames.Count; c++)
                    {
                        if ((String)suppress_count.classifierNames[c] == (String)temp_group.classifierNames[b])
                        {
                            tempCount = (Int64)suppress_count.classifierCount[c];
                            tempCount += tempClassifierCount_Suppressed;
                            suppress_count.classifierCount[c] = tempCount;
                            suppress_count.groupCount += tempClassifierCount_Suppressed;

                            suppressFlag = true;
                            break;

                        }//if ((String) suppress_count.classifierNames[c] == (String) temp_group.classifierNames[c])

                    }//for (int c = 0; c < suppress_count.classifierNames.Count; c++)

                    if (!suppressFlag)
                    {
                        suppress_count.classifierNames.Add(temp_group.classifierNames[b]);
                        suppress_count.classifierCount.Add(tempClassifierCount_Suppressed);
                        suppress_count.groupCount += tempClassifierCount_Suppressed;
                    }
                                                                                
                    tempClassifierCount = (Int64)temp_group_suppressed.classifierCount[b];
                    suppressFlag = false;

                    for (int d = 0; d < temp_targetAttribute.classifierNames.Count; d++)
                    {
                        if ((String) temp_targetAttribute.classifierNames[d] == (String) temp_group.classifierNames[b])
                        {
                            tempCount = (Int64)temp_targetAttribute.classifierCount[d];
                            tempCount += tempClassifierCount;
                            temp_targetAttribute.classifierCount[d] = tempCount;
                            temp_targetAttribute.groupCount += tempClassifierCount;

                            suppressFlag = true;
                            break;
                        }

                    }//for (int d = 0; d < temp_targetAttribute.classifierNames.Count; d++)

                    if (!suppressFlag)
                    {
                        temp_targetAttribute.classifierNames.Add(temp_group.classifierNames[b]);
                        temp_targetAttribute.classifierCount.Add(tempClassifierCount);
                        temp_targetAttribute.groupCount += tempClassifierCount;
                    }

                    suppressGroupCount += tempClassifierCount_Suppressed;

                }//for (int b = 0; b < temp_group.classifierNames.Count; b++)
                                
                temp_suppressedLevel.Add(temp_group_suppressed);

            }//for(int a = 0; a < class_High_Low_Count.levelGroups.Count; a++)

            suppress_count.groupCount = suppressGroupCount;

            if (classNum == 3 || classNum == 16)
                logValues.WriteLine("QI Group = " + suppress_count.groupName + ", Count = " + suppress_count.groupCount);

            

            suppressGroupCount = 0;

            for (int c = 0; c < temp_targetAttribute.classifierNames.Count; c++)
                suppressGroupCount += (Int64) temp_targetAttribute.classifierCount[c];

            temp_targetAttribute.groupCount = suppressGroupCount;
            
            temp_targetAttributeGroup.Add(temp_targetAttribute);

            for(int b = 0; b < Suppress_class_High_Low_Count.levelGroups.Count; b++)
            {
                temp_group = (Group)Suppress_class_High_Low_Count.levelGroups[b];

                for (int d = 0; d < temp_group.classifierNames.Count; d++)
                {
                    suppressGroupCount = (Int64)temp_group.classifierCount[d];

                    infoGainRatio += suppressGroupCount == 0 ? 0 : (double)suppressGroupCount / (record_count - suppress_count.groupCount) * 
                                                                            Math.Log((double)suppressGroupCount / temp_group.groupCount, 2);

                }//for (int d = 0; d < temp_group.classifierNames.Count; d++)

            }//for(int b = 0; b < Suppress_class_High_Low_Count.levelGroups.Count; b++)

            infoGainRatio = -infoGainRatio;

            Level temp_level = (Level)Levels_tree[0];
            temp_level = (Level)Levels_tree[temp_level.maxTradeoffLevel];

            infoGainRatio = (double)CalculateTargetAttributeEntropy(temp_targetAttributeGroup) - infoGainRatio;
            infoGainRatio = (double)infoGainRatio / CalculateLevelEntropy(ref temp_suppressedLevel);

            infoGainRatio = Math.Abs(infoGainRatio - temp_level.levelMutualInformationNormalized);

            //logValues.WriteLine("levelMutualInformationNormalized " + temp_level.levelMutualInformationNormalized);

            return infoGainRatio;

        }//public double CalculateInformationGainRatio()

        //Kullback-Leibler Divergence 
        public void CalculateDistributionDistance(int qiIndex, ref SuppressedClass suppressClass, Level class_High_Low_Count, Level Suppress_class_High_Low_Count, ArrayList Levels_tree)
        {
            //qiNames[qiIndex]

            double distributionDistance = 0;

            Group temp_group;
            Group temp_group_suppressed;

            double tempA = 0, tempB = 0;
            Int64 classifierCount = 0, classifierCount_Suppressed = 0, tempDifference = 0;
            bool suppressFlag = false;

            ArrayList temp_suppressedLevel = new ArrayList();
            Group suppress_count = new Group();
            suppress_count.groupName = "?";
            suppress_count.classifierCount = new ArrayList();
            suppress_count.classifierNames = new ArrayList();

            for(int a = 0; a < class_High_Low_Count.levelGroups.Count; a++)
            {
                temp_group = (Group)class_High_Low_Count.levelGroups[a];
                temp_group_suppressed = (Group)Suppress_class_High_Low_Count.levelGroups[a];

                //==============================================================================
                //if (temp_group.groupCount == 0)
                //{
                //    if (trainingFlag == false && testFlag == false)
                //        tempA = (double)0.000001 / tuples.Length;
                //    if (trainingFlag == true && testFlag == false)
                //        tempA = (double)0.000001 / trainingTuples.Length;
                //    if (trainingFlag == false && testFlag == true)
                //        tempA = (double)0.000001 / testTuples.Length;
                //}
                //else
                //{
                //    if (trainingFlag == false && testFlag == false)
                //        tempA = (double)temp_group.groupCount / tuples.Length;
                //    if (trainingFlag == true && testFlag == false)
                //        tempA = (double)temp_group.groupCount / trainingTuples.Length;
                //    if (trainingFlag == false && testFlag == true)
                //        tempA = (double)temp_group.groupCount / testTuples.Length;
                //}

                //if (temp_group_suppressed.groupCount == 0)
                //{
                //    if (trainingFlag == false && testFlag == false)
                //        tempB = (double)Math.Log((double)0.000001 / tuples.Length, 2);
                //    if (trainingFlag == true && testFlag == false)
                //        tempB = (double)Math.Log((double)0.000001 / trainingTuples.Length, 2);
                //    if (trainingFlag == false && testFlag == true)
                //        tempB = (double)Math.Log((double)0.000001 / testTuples.Length, 2);
                //}
                //else
                //{
                //    if (trainingFlag == false && testFlag == false)
                //        tempB = (double)Math.Log((double)temp_group_suppressed.groupCount / tuples.Length, 2);
                //    if (trainingFlag == true && testFlag == false)
                //        tempB = (double)Math.Log((double)temp_group_suppressed.groupCount / trainingTuples.Length, 2);
                //    if (trainingFlag == false && testFlag == true)
                //        tempB = (double)Math.Log((double)temp_group_suppressed.groupCount / testTuples.Length, 2);
                //}
                //====================================================================================

                if (temp_group.groupCount == 0)
                {
                    tempA = 0; 
                    tempB = 0;
                }
                else if (temp_group.groupCount > 0 && temp_group_suppressed.groupCount > 0)
                {
                    if (trainingFlag == false && testFlag == false)
                        tempA = (double)temp_group.groupCount / tuples.Length;
                    if (trainingFlag == true && testFlag == false)
                        tempA = (double)temp_group.groupCount / trainingTuples.Length;
                    if (trainingFlag == false && testFlag == true)
                        tempA = (double)temp_group.groupCount / testTuples.Length;

                    if (trainingFlag == false && testFlag == false)
                        tempB = (double)Math.Log((double)temp_group_suppressed.groupCount / tuples.Length, 2);
                    if (trainingFlag == true && testFlag == false)
                        tempB = (double)Math.Log((double)temp_group_suppressed.groupCount / trainingTuples.Length, 2);
                    if (trainingFlag == false && testFlag == true)
                        tempB = (double)Math.Log((double)temp_group_suppressed.groupCount / testTuples.Length, 2);

                }
                else if (temp_group.groupCount > 0 && temp_group_suppressed.groupCount == 0)
                {
                    if (trainingFlag == false && testFlag == false)
                        tempA = (double)temp_group.groupCount / tuples.Length;
                    if (trainingFlag == true && testFlag == false)
                        tempA = (double)temp_group.groupCount / trainingTuples.Length;
                    if (trainingFlag == false && testFlag == true)
                        tempA = (double)temp_group.groupCount / testTuples.Length;

                    if (trainingFlag == false && testFlag == false)
                        tempB = (double)Math.Log((double)0.49 / tuples.Length, 2);
                    if (trainingFlag == true && testFlag == false)
                        tempB = (double)Math.Log((double)0.49 / trainingTuples.Length, 2);
                    if (trainingFlag == false && testFlag == true)
                        tempB = (double)Math.Log((double)0.49 / testTuples.Length, 2);
                }

                distributionDistance = distributionDistance + (tempA * tempB);

                temp_suppressedLevel.Add(temp_group_suppressed);


                for (int b = 0; b < temp_group.classifierNames.Count; b++)
                {
                    classifierCount = (Int64)temp_group.classifierCount[b];
                    classifierCount_Suppressed = (Int64)temp_group_suppressed.classifierCount[b];

                    tempDifference = classifierCount - classifierCount_Suppressed;

                    suppressFlag = false;
                    for (int c = 0; c < suppress_count.classifierNames.Count; c++)
                    {
                        if ((String)suppress_count.classifierNames[c] == (String)temp_group.classifierNames[b])
                        {   
                            classifierCount = (Int64)suppress_count.classifierCount[c];
                            classifierCount += tempDifference;
                            suppress_count.classifierCount[c] = classifierCount;

                            suppress_count.groupCount += tempDifference;

                            suppressFlag = true;
                            break;
                        }

                    }//for (int c = 0; c < suppress_count.classifierNames.Count; c++)

                    if (!suppressFlag)
                    {
                        suppress_count.classifierCount.Add(tempDifference);
                        suppress_count.classifierNames.Add(temp_group.classifierNames[b]);
                        suppress_count.groupCount += tempDifference;
                    }

                }//for (int b = 0; b < temp_group.classifierNames.Count; b++)

            }//for(int a = 0; a < class_High_Low_Count.levelGroups.Count; a++)

            temp_suppressedLevel.Add(suppress_count);

            //actually makes no difference
            //distributionDistance += (0/record_count * Math.Log((suppress/record_count))); 0 = number of ? in generalized data

            distributionDistance = -distributionDistance;


            Level temp_level = (Level)Levels_tree[0];
            temp_level = (Level)Levels_tree[temp_level.maxTradeoffLevel];


            suppressClass.distribution_distance = distributionDistance - temp_level.levelEntropy;

            //logValues.WriteLine("Inside CalculateDistributionDistance, QI: " + );
            
            suppressClass.distribution_distance_normalized = suppressClass.distribution_distance / temp_level.levelEntropy;


        }//public void CalculateDistributionDistance()

        public void CreateOrigionalWekaDataSet()
        {
            int level = -2;
            int nodeCount;
            int levelCount = -2, levelNodesCount = 0;
            int tuplesCount = 0;
            
            
            StreamWriter origional_Data_Weka;

            if (trainingFlag == true)
            {
                origional_Data_Weka = new StreamWriter("output\\temp\\origional_weka_training.arff");
                tuplesCount = trainingTuples.Length;
            }
            else if (testFlag == true)
            {
                origional_Data_Weka = new StreamWriter("output\\temp\\origional_weka_test.arff");
                tuplesCount = testTuples.Length;
            }
            else
            {
                origional_Data_Weka = new StreamWriter("output\\origional_weka.arff");
                tuplesCount = tuples.Length;
            }

            origional_Data_Weka.AutoFlush = true;

            origional_Data_Weka.WriteLine("%\t Origional Data File Created at " + DateTime.Now.ToString());

            origional_Data_Weka.WriteLine("");
            origional_Data_Weka.WriteLine("");
            origional_Data_Weka.WriteLine("");

            origional_Data_Weka.WriteLine("@relation Origional_dataset_count_"+ tuples.Length);

           
            level = -2;// -1 because the name of QI is on top level in generalization tree
            for (int b = 0; b < QI_GeneralizationTree.Length; b++)
            {
                nodeCount = QI_GeneralizationTree[b].Nodes.Count;
                
                //sids
                origional_Data_Weka.Write("@attribute " + qids[b].InnerText + " {");

                levelCount = -2;
                levelNodesCount = 0;
                GetLevelNodesCount(QI_GeneralizationTree[b].Nodes, (QIs_Levels[b].Count - 1), ref levelCount, ref levelNodesCount);
                
                level = -2;
                WriteWholeGeneralizationTree(ref origional_Data_Weka, QI_GeneralizationTree[b].Nodes, (QIs_Levels[b].Count - 1), ref level);

                origional_Data_Weka.BaseStream.Position = origional_Data_Weka.BaseStream.Position - 1;
                origional_Data_Weka.WriteLine("}");

            }//for (int b = 0; a < QI_GeneralizationTree.Length; a++)
            
            origional_Data_Weka.WriteLine("");

            origional_Data_Weka.WriteLine("@data");

            for (int d = 0; d < tuplesCount; d++)
            {
                if(trainingFlag == true)
                    origional_Data_Weka.WriteLine(trainingTuples[d]);
                else if(testFlag == true)
                    origional_Data_Weka.WriteLine(testTuples[d]);
                else
                    origional_Data_Weka.WriteLine(tuples[d]);
            }

            origional_Data_Weka.Close();

        }//public void CreateOrigionalWekaDataSet()
        
        public bool WriteWholeGeneralizationTree(ref StreamWriter origional_Data_Weka, TreeNodeCollection QI_Parent, int level, ref int levelCount)
        {
            bool returnValue = false;
            int nodeCount = QI_Parent.Count;

            if (nodeCount == 0)
            {
                if (levelCount < level)
                    return true;

                return false;
            }

            levelCount++;
            for (int a = 0; a < nodeCount; a++)
            {
                //MessageBox.Show("Level: " + level + " Value: " + QI_Parent[a].Text);
                if (levelCount == level)
                    origional_Data_Weka.Write(QI_Parent[a].Text+ ",");
                
                returnValue = WriteWholeGeneralizationTree(ref origional_Data_Weka, QI_Parent[a].Nodes, level, ref levelCount);

                if (returnValue)
                    origional_Data_Weka.Write(QI_Parent[a].Text + ",");
                
            }
            levelCount--;

            return false;

        }//public void WriteWholeGeneralizationTree(ref Group group, int level)

        public void InitializeSuppressWekaDataSet(ref StreamWriter suppress_data, ref StreamWriter suppress_writer_SVM, int suppressClass)
        {
            if (suppress_data != null)
            {
                suppress_data.Flush();
                suppress_data.Close();
                suppress_data = null;
            }

            if (suppress_writer_SVM != null)
            {
                suppress_writer_SVM.Flush();
                suppress_writer_SVM.Close();
                suppress_writer_SVM = null;
            }

            if (trainingFlag == true)
            {
                suppress_data = new StreamWriter("output\\temp\\suppress_class_training_" + suppressClass.ToString() + ".arff");

                if (classifierFlag[SVM] == "1")
                {
                    suppress_writer_SVM = new StreamWriter("output\\temp\\suppress_class_training_SVM_" + suppressClass.ToString() + ".arff");
                    suppress_writer_SVM.AutoFlush = true;
                }
            }
            else
                suppress_data = new StreamWriter("output\\suppress_class_" + suppressClass.ToString() + ".arff");
            
            
            suppress_data.AutoFlush = true;

            suppress_data.WriteLine("%\t Generalized Data File Created at " + DateTime.Now.ToString());

            suppress_data.WriteLine("");
            suppress_data.WriteLine("");
            suppress_data.WriteLine("");

            suppress_data.WriteLine("@relation generalized_suppress_class_count_"+suppressClass.ToString());

            if (suppress_writer_SVM != null)
            {
                suppress_writer_SVM.WriteLine("%\t Generalized Data File Created at " + DateTime.Now.ToString());

                suppress_writer_SVM.WriteLine("");
                suppress_writer_SVM.WriteLine("");
                suppress_writer_SVM.WriteLine("");

                suppress_writer_SVM.WriteLine("@relation generalized_suppress_class_count_" + suppressClass.ToString());
            }

            for (int b = 0; b < totalQIs; b++)
            {
                //sids
                suppress_data.Write("@attribute " + qids[b].InnerText + " {");

                if(suppress_writer_SVM != null)
                    suppress_writer_SVM.Write("@attribute " + qids[b].InnerText + " {");

                WriteGeneralizationTree(ref suppress_data, ref suppress_writer_SVM, maximumTradeoffLevels[b], QIs_Levels[b]);

                suppress_data.BaseStream.Position = suppress_data.BaseStream.Position - 1;
                suppress_data.WriteLine("}");

                if (suppress_writer_SVM != null)
                {
                    suppress_writer_SVM.BaseStream.Position = suppress_writer_SVM.BaseStream.Position - 1;
                    suppress_writer_SVM.WriteLine("}");
                }

            }//for (int b = 0; a < QI_GeneralizationTree.Length; a++)

            suppress_data.WriteLine("");
            suppress_data.WriteLine("@data");

            if (suppress_writer_SVM != null)
            {
                suppress_writer_SVM.WriteLine("");
                suppress_writer_SVM.WriteLine("@data");
            }

        }//public void InitializeSuppressWekaDataSet(StreamWriter suppress_data, int suppressClass)

        public void WriteSuppressWekaDataSet(ref StreamWriter suppress_data, ref StreamWriter suppress_writer_SVM, Hashtable class_tuples, int classStart, int classEnd)
        {
            ArrayList temp_class;
            DistinctTuple temp_tuple;

            for (int a = 1; a <= class_tuples.Count; a++)
            {
                temp_class = (ArrayList)class_tuples[a];

                for (int b = 0; b < temp_class.Count; b++)
                {
                    temp_tuple = (DistinctTuple)temp_class[b];

                    for (int c = 0; c < temp_tuple.tuple.Count; c++)
                    {
                        /////////////////////////////////////////////////////
                        //if (temp_tuple.tuple[c].ToString().IndexOf("?") >= 0)     //uncomment only for SVM data set because SVM donot support ?
                        //{
                        //    temp_tuple.tuple[c] = temp_tuple.tuple[c].ToString().Replace("?","Any");
                        //    //continue;
                        //}
                        //////////////////////////////////////////////////////

                        suppress_data.WriteLine(temp_tuple.tuple[c].ToString());

                        if(suppress_writer_SVM != null)
                            suppress_writer_SVM.WriteLine((temp_tuple.tuple[c].ToString()).Replace("?","Any"));
                    }
                    
                }//for (int b = 0; b < temp_class.Count; b++)

            }//for (int a = classStart; a <= classEnd; a++)

        }//public void WriteSuppressWekaDataSet(ref StreamWriter suppress_data, Hashtable class_tuples, int classStart, int classEnd)

        public void CreateLevels(ref ArrayList levels, int QI_number)
        {
            ArrayList groups;
            Level level;
            int levelNodesCount = 0;
            int levelCount = -2;

            
            logValues.WriteLine("");
            for (int a = 0; a < levels.Capacity; a++)
            {
                levelCount = -2;
                levelNodesCount = 0;
                GetLevelNodesCount(QI_GeneralizationTree[QI_number].Nodes, a, ref levelCount, ref levelNodesCount);
                groups = new ArrayList(levelNodesCount);

                levelCount = -2;
                levelNodesCount = 0;
                ReadLevelValues(QI_GeneralizationTree[QI_number].Nodes, a, ref levelCount, ref levelNodesCount, ref groups);
                
                EnumerateGroups(ref groups, QI_number);

                level = new Level();
                //calculate level entroyp
                level.levelEntropy = CalculateLevelEntropy(ref groups);

                if(trainingFlag == false && testFlag == false)
                    logValues.WriteLine(qids[QI_number].InnerText + " Entropy level " + a + " = " + level.levelEntropy.ToString());
                else if (trainingFlag = true && testFlag == false)
                    logValues.WriteLine("Training_QI " + qids[QI_number].InnerText + " Entropy level " + a + " = " + level.levelEntropy.ToString()); //0.13
                
                level.levelGroups = groups;

                if (sensitiveFlag[QI_number] == 1)
                    level.maxTradeoffLevel = levels.Capacity - 1;
                
                levels.Add(level);
            
            }//for (int a = 0; a < levels.Capacity; a++)

            if (testFlag == true || sensitiveFlag[QI_number] == 1)
                return;

            Level levelOne = (Level)levels[0];
            Level levelEnd = (Level)levels[levels.Count - 1];

            logValues.WriteLine("");
            levelOne.targetAttributeEntropy = CalculateTargetAttributeEntropy(levelOne.levelGroups);

            if (trainingFlag == false && testFlag == false)
                logValues.WriteLine("Targer Attribute Entropy " + levelOne.targetAttributeEntropy); //age 0.8078
            else if (trainingFlag = true && testFlag == false)
                logValues.WriteLine("Training Data Targer Attribute Entropy " + levelOne.targetAttributeEntropy); //0.8085
            
            logValues.WriteLine("");
            //calculate level specification loss
            for (int b = 0; b < levels.Count; b++)
            {
                level = (Level)levels[b];

                level.targetAttributeEntropy = levelOne.targetAttributeEntropy;
                
                level.levelSpecLoss = CalculateLevelSpecLoss(level.levelEntropy, levelEnd.levelEntropy);    
                level.levelMutualInformation = CalculateLevelMutualInformation(level.levelGroups, level.targetAttributeEntropy);
                level.levelMutualInformationNormalized = level.levelEntropy != 0 ? (double)level.levelMutualInformation / level.levelEntropy : 0; 
                
                levels[b] = level;

            }//for (int b = 0; b < levels.Capacity; b++)

            logValues.WriteLine("");
            for (int c = 0; c < levels.Count; c++)
            {
                level = (Level)levels[c];

                if (trainingFlag == false && testFlag == false)
                    logValues.WriteLine(qids[QI_number].InnerText + " Specification Loss level " + c + " = " + level.levelSpecLoss.ToString());
                else if (trainingFlag = true && testFlag == false)
                    logValues.WriteLine("Training_QI " + qids[QI_number].InnerText + " Specification Loss level " + c + " = " + level.levelSpecLoss.ToString());


            }//for (int c = 0; c < levels.Count; c++)

            logValues.WriteLine("");
            for (int d = 0; d < levels.Count; d++)
            {
                level = (Level)levels[d];

                if (trainingFlag == false && testFlag == false)
                    logValues.WriteLine(qids[QI_number].InnerText + " Mutual Information level " + d + " = " + level.levelMutualInformation.ToString());
                else if (trainingFlag = true && testFlag == false)
                    logValues.WriteLine("Training_QI " + qids[QI_number].InnerText + " Mutual Information level " + d + " = " + level.levelMutualInformation.ToString());

            }//for (int d = 0; d < levels.Count; d++)

            logValues.WriteLine("");
            for (int e = 0; e < levels.Count; e++)
            {
                level = (Level)levels[e];

                if (trainingFlag == false && testFlag == false)
                    logValues.WriteLine(qids[QI_number].InnerText + " Mutual Information Normalized level " + e + " = " + level.levelMutualInformationNormalized.ToString());
                else if (trainingFlag = true && testFlag == false)
                    logValues.WriteLine("Training_QI " + qids[QI_number].InnerText + " Mutual Information Normalized level " + e + " = " + level.levelMutualInformationNormalized.ToString());


            }//for (int e = 0; e < levels.Count; e++)


            //double maxlevelWorkclassDifferenceMutualInformationNormalized = CalculatemaxlevelDifferenceMutualInformationNormalized(workclass_Levels_Tree);
            double maxLevelDifferenceMutualInformationNormalized = CalculatemaxlevelDifferenceMutualInformationNormalized(levels); 

            logValues.WriteLine("");
            levelEnd = (Level)levels[levels.Count - 1];
            
            for (int f = 0; f < levels.Count; f++)
            {
                level = (Level) levels[f];
                level.maxLevelDifferenceMutualInformationNormalized = maxLevelDifferenceMutualInformationNormalized; 
                level.levelClassGain = CalculateClassGain(level, levelEnd.levelMutualInformationNormalized, maxLevelDifferenceMutualInformationNormalized); 

                if (trainingFlag == false && testFlag == false)
                    logValues.WriteLine(qids[QI_number].InnerText + " Class Gain level " + f + " = " + level.levelClassGain.ToString());
                else if (trainingFlag = true && testFlag == false)
                    logValues.WriteLine("Training_QI " + qids[QI_number].InnerText + " Class Gain level " + f + " = " + level.levelClassGain.ToString());



                //TEMP CHANGE 01-06-2009
                //level.levelTradeoff = (double)0.5 * level.levelClassGain - 0.5 * level.levelSpecLoss;
                level.levelTradeoff = level.levelClassGain;

                levels[f] = level;

            }//for (int f = 0; f < levels.Count; f++)

            logValues.WriteLine("");
            logValues.WriteLine("TEMP CHANGE ON 01-06-2009 level.levelTradeoff = level.levelClassGain");
            for (int g = 0; g < levels.Count; g++)
            {
                level = (Level)levels[g];

                if (trainingFlag == false && testFlag == false)
                    logValues.WriteLine(qids[QI_number].InnerText + " Tradeoff level " + g + " = " + level.levelTradeoff.ToString());
                else if (trainingFlag = true && testFlag == false)
                    logValues.WriteLine("Training_QI " + qids[QI_number].InnerText + " Tradeoff level " + g + " = " + level.levelTradeoff.ToString());

                
            }//for (int g = 0; g < levels.Count; g++)

            int maximumWorkclassTradeoffLevel = CalculateMaximumTradeoffLevel(ref levels, ref sensitiveFlag);
            

        }//public void CreateLevels(ref ArrayList levels)

        public double CalculatemaxlevelDifferenceMutualInformationNormalized(ArrayList levels)
        {
            double maxlevelDifferenceMutualInformationNormalized = 0;
            double tempDifference = 0;

            Level tempLevel_Lowest = (Level) levels[levels.Count - 1]; // levels_Tree[levels_Tree.Count.ToString()];
            Level tempLevel_Current = (Level) levels[0]; //levels_Tree["1"];

            maxlevelDifferenceMutualInformationNormalized = tempLevel_Current.levelMutualInformationNormalized - tempLevel_Lowest.levelMutualInformationNormalized;

            for (int a = 1; a < levels.Count; a++)
            {
                tempLevel_Current = (Level)levels[a];

                tempDifference = tempLevel_Current.levelMutualInformationNormalized - tempLevel_Lowest.levelMutualInformationNormalized;

                if (tempDifference > maxlevelDifferenceMutualInformationNormalized)
                    maxlevelDifferenceMutualInformationNormalized = tempDifference;

            }

            return maxlevelDifferenceMutualInformationNormalized;

        }//public double CalculatemaxlevelDifferenceMutualInformationNormalized(ArrayList levels)

        public void EnumerateGroups(ref ArrayList groups, int QI_number)
        {
            Group tempGroup;
            int firstIndex = 0, lastIndex = 0, loopCount = 0;
            int foundLevel = 0;

            String line = "", token = "";

            if (trainingFlag == true)
                loopCount = trainingTuples.Length;
            else if (testFlag == true)
                loopCount = testTuples.Length;
            else
                loopCount = tuples.Length;
            

            for (int a = 0; a < loopCount; a++)
            {
                if (trainingFlag == true)
                    line = trainingTuples[a];
                else if (testFlag == true)
                    line = testTuples[a];
                else
                    line = tuples[a];

                if (line == null)
                    break;

                if (QI_number == 0)
                {
                    lastIndex = line.IndexOf(',');
                    token = line.Substring(firstIndex, lastIndex);
                }
                else
                {
                    firstIndex = line.IndexOf(',');

                    for (int b = 0; b < QI_number-1; b++)
                    {
                        firstIndex++;
                        firstIndex = line.IndexOf(',', firstIndex);
                    }

                    firstIndex++;
                    lastIndex = line.IndexOf(',', firstIndex);

                    if(lastIndex > 0)
                        token = line.Substring(firstIndex, lastIndex - firstIndex);
                    else
                        token = line.Substring(firstIndex);
                }// else of if (QI_number == 0)

                

                if (groups.Count == 1)   // group only have ONE node having top-most level i.e. 'Any'
                {
                    tempGroup = (Group) groups[0];
                    tempGroup.groupCount++;
                    CountHighLow(line, ref tempGroup);
                    groups[0] = tempGroup;
                }
                else
                {
                    for (int b = 0; b < groups.Count; b++)
                    {
                        tempGroup = (Group) groups[b];

                        foundLevel = 0;
                        if ((tempGroup.groupTree.Text == token) || (FoundInGroupTree(tempGroup.groupTree.Nodes, token,ref foundLevel)))
                        {
                            tempGroup.groupCount++;
                            CountHighLow(line, ref tempGroup);
                            groups[b] = tempGroup;
                            break;
                        }


                    }//for (int b = 0; b < group.Length; b++)

                }//else of if (group.Length == 1)

            }//for (int a = 0; a < tuples.Length; a++)

        }//public void EnumerateGroups(ref Group[] group)

        public bool FoundInGroupTree(TreeNodeCollection groupTree, String token, ref int foundLevel)
        {
            bool returnValue = false;
            int nodeCount = groupTree.Count;

            if (nodeCount == 0)
            {
                return false;
            }

            foundLevel++;
            for (int a = 0; a < nodeCount; a++)
            {
                if (groupTree[a].Text == token)
                    return true;

                
                returnValue = FoundInGroupTree(groupTree[a].Nodes, token, ref foundLevel);
                

                if (returnValue == true)
                    break;
            }
            
            return returnValue;

        }//public bool FoundInGroupTree(TreeNodeCollection groupTree, String token)

        public void GetLevelNodesCount(TreeNodeCollection QI_Parent, int level, ref int levelCount, ref int levelNodesCount)
        {
            
            int nodeCount = QI_Parent.Count;

            if (nodeCount == 0)
            {
                if (levelCount < level)
                    levelNodesCount++;

                return;
            }

            levelCount++;
            for (int a = 0; a < nodeCount; a++)
            {
                //MessageBox.Show("Level: " + level + " Value: " + QI_Parent[a].Text);
                if (levelCount == level)
                    levelNodesCount++;

                GetLevelNodesCount(QI_Parent[a].Nodes, level, ref levelCount, ref levelNodesCount);
            }
            levelCount--;

            return;

        }//void ReadWholeGeneralizationTree(TreeNodeCollection QI_Parent)

        public bool ReadLevelValues(TreeNodeCollection QI_Parent, int level, ref int levelCount, ref int levelNodesCount, ref ArrayList groups)
        {
            Group tempGroup;
            bool returnValue = false;
            int nodeCount = QI_Parent.Count;

            if (nodeCount == 0)
            {
                if (levelCount < level)
                    return true;
                

                return false;
            }

            levelCount++;
            for (int a = 0; a < nodeCount; a++)
            {
                //MessageBox.Show("Level: " + level + " Value: " + QI_Parent[a].Text);
                if (levelCount == level)
                {
                    tempGroup = new Group();
                    tempGroup.groupName = QI_Parent[a].Text;
                    tempGroup.groupTree = QI_Parent[a];

                    groups.Add(tempGroup);

                    //group[levelNodesCount].groupName = QI_Parent[a].Text;
                    //group[levelNodesCount].groupTree = QI_Parent[a];
                    levelNodesCount++;
                }

                returnValue = ReadLevelValues(QI_Parent[a].Nodes, level, ref levelCount, ref levelNodesCount, ref groups);

                if (returnValue)
                {
                    tempGroup = new Group();
                    tempGroup.groupName = QI_Parent[a].Text;
                    tempGroup.groupTree = QI_Parent[a];

                    groups.Add(tempGroup);
                    
                    levelNodesCount++;
                }

            }
            levelCount--;

            return false;


        }//public void ReadGroupSubTreeValues(ref Group group, int level)

        SymbolType RandomlySelectSymbol(ref int symbolIndex)
        {
            //Thread.Sleep(1000);
            //Random randomSymbol = new Random(DateTime.Now.Millisecond);
            //int symbolIndex = randomSymbol.Next(10);

            Symbol symbol = new Symbol();

            if (symbolIndex > 8)
                symbolIndex = 1;
                        
            switch (symbolIndex)
            {
                case 1: return SymbolType.Circle;
                case 2: return SymbolType.Diamond;
                //case 3: return SymbolType.HDash;
                case 3: return SymbolType.Plus;
                case 4: return SymbolType.Square;
                case 5: return SymbolType.Star;
                case 6: return SymbolType.Triangle;
                case 7: return SymbolType.TriangleDown;
                //case 8: return SymbolType.VDash;
                case 8: return SymbolType.XCross;
            }

            return SymbolType.Default;

        }//SymbolType RandomlySelectSymbol()

        ////////////////////////////////////////////////////////////
        //Graphs

        /*  Age             Magenta         Star
         *  Workclass       Aqua            XCross
         *  Education-Num   LimeGreen       Plus
         *  Marital Status  HotPink         Diamond
         *  Occuption       Maroon          VDash
         *  Race            Green           Circle
         *  Sex             Red             Triangle
         *  Native Country  Blue            Square
         */

        public void CreateGraphAGain(ZedGraphControl treeGraph)
        {
            Level tempLevel;
            GraphPane treeGraphPane = treeGraph.GraphPane;

            //titles of the graph
            treeGraphPane.Title.Text = "Level Normalized Mutual Information";
            treeGraphPane.XAxis.Title.Text = "Levels";
            treeGraphPane.YAxis.Title.Text = "Classification Gain";

            this.logValues.WriteLine("Level Normalized Mutual Information");

            treeGraphPane.XAxis.Scale.IsReverse = true;
            // treeGraphPane.YAxis.Scale.IsReverse = true;

            int generalizedQIsCount = 0;
            int pairCount = 0;

            for(int c = 0; c < totalQIs; c++)
                if(sensitiveFlag[c] == 0)
                    generalizedQIsCount++;

            PointPairList[] pairClassGain = new PointPairList[generalizedQIsCount];

            for (int d = 0; d < pairClassGain.Length; d++)
                pairClassGain[d] = new PointPairList();

            ArrayList QI_Levels;
            int QIs = 0;



            Random random_RGB = new Random(DateTime.Now.Millisecond);
            Color[] color = new Color[generalizedQIsCount];

            int symbolIndex = 0;

            
            for (int a = 0; a < totalQIs; a++)
            {
                if (sensitiveFlag[a] == 1)
                    continue;

                this.logValues.WriteLine("");
                this.logValues.WriteLine("QI = " + qiNames[a]);
                this.logValues.WriteLine("QI Level, Level Classification Gain");

                QI_Levels = QIs_Levels[a];
                QIs = QI_Levels.Count;

                for (int b = 0; b < QI_Levels.Count; b++)
                {
                    tempLevel = (Level)QI_Levels[b];

                    if (tempLevel.levelClassGain > -0.5)
                    {
                        pairClassGain[pairCount].Add(QIs, tempLevel.levelClassGain);
                        this.logValues.WriteLine(QIs.ToString() + " , " + tempLevel.levelClassGain);
                    }
                    else
                    {
                        pairClassGain[pairCount].Add(QIs, -0.5);
                        this.logValues.WriteLine(QIs.ToString() + " , -0.5");
                    }

                    QIs--;
                    
                }//for (int b = 0; b < QIs_Levels[a].Count; b++)

                color[pairCount] = Color.FromArgb(255,random_RGB.Next(0,255),random_RGB.Next(0,255),random_RGB.Next(0,255));

                for (int d = 0; d < color.Length; d++)
                {
                    if ((d != pairCount) && (color[pairCount] == color[d]))
                    {
                        color[pairCount] = Color.FromArgb(255, random_RGB.Next(0, 255), random_RGB.Next(0, 255), random_RGB.Next(0, 255));
                        d = 0;
                    }
                }

                symbolIndex++;
                LineItem curveClassGain = treeGraphPane.AddCurve(qiNames[a], pairClassGain[pairCount], color[pairCount], RandomlySelectSymbol(ref symbolIndex));

                pairCount++;

            }//for (int a = 0; a < totalQIs; a++)

            this.logValues.WriteLine("");

            treeGraph.AxisChange();

        } //public void CreateGraphAGain(ZedGraphControl treeGraph)

        public void CreateGraphTradeoff(ZedGraphControl treeGraph)
        {
            Level tempLevel;
            GraphPane treeGraphPane = treeGraph.GraphPane;

            //titles of the graph
            treeGraphPane.Title.Text = "Level Trade-off";
            treeGraphPane.XAxis.Title.Text = "Levels";
            treeGraphPane.YAxis.Title.Text = "Trade-off";

            treeGraphPane.XAxis.Scale.IsReverse = true;
            // treeGraphPane.YAxis.Scale.IsReverse = true;

            int generalizedQIsCount = 0;
            int pairCount = 0;

            for (int c = 0; c < totalQIs; c++)
                if (sensitiveFlag[c] == 0)
                    generalizedQIsCount++;

            PointPairList[] pairTradeOff = new PointPairList[generalizedQIsCount];

            for (int d = 0; d < pairTradeOff.Length; d++)
                pairTradeOff[d] = new PointPairList();


            ArrayList QI_Levels;
            int QIs = 0;

            Random random_RGB = new Random(DateTime.Now.Millisecond);
            Color[] color = new Color[generalizedQIsCount];

            int symbolIndex = 0;
            for (int a = 0; a < totalQIs; a++)
            {
                if (sensitiveFlag[a] == 1)
                    continue;

                QI_Levels = QIs_Levels[a];
                QIs = QI_Levels.Count;

                for (int b = 0; b < QI_Levels.Count; b++)
                {
                    tempLevel = (Level)QI_Levels[b];

                    if (tempLevel.levelTradeoff > -0.5)
                        pairTradeOff[pairCount].Add(QIs, tempLevel.levelTradeoff);
                    else
                        pairTradeOff[pairCount].Add(QIs, -0.5);

                    QIs--;

                }//for (int b = 0; b < QIs_Levels[a].Count; b++)

                color[pairCount] = Color.FromArgb(255, random_RGB.Next(0, 255), random_RGB.Next(0, 255), random_RGB.Next(0, 255));

                for (int d = 0; d < color.Length; d++)
                {
                    if ((d != pairCount) && (color[pairCount] == color[d]))
                    {
                        color[pairCount] = Color.FromArgb(255, random_RGB.Next(0, 255), random_RGB.Next(0, 255), random_RGB.Next(0, 255));
                        d = 0;
                    }
                }

                symbolIndex++;
                LineItem curveClassGain = treeGraphPane.AddCurve(qiNames[a], pairTradeOff[pairCount], color[pairCount], RandomlySelectSymbol(ref symbolIndex));

                pairCount++;

            }//for (int a = 0; a < totalQIs; a++)

            treeGraph.AxisChange();

        } //public void CreateGraphAGain_Tradeoff(ZedGraphControl treeGraph)

        public void CreateGraphEntropy(ZedGraphControl treeGraph)
        {
            Level tempLevel;
            GraphPane treeGraphPane = treeGraph.GraphPane;


            //titles of the graph
            treeGraphPane.Title.Text = "Level Entropy";
            treeGraphPane.XAxis.Title.Text = "Levels";
            treeGraphPane.YAxis.Title.Text = "Entropy";

            treeGraphPane.XAxis.Scale.IsReverse = true;
            // treeGraphPane.YAxis.Scale.IsReverse = true;

            int generalizedQIsCount = 0;
            int pairCount = 0;

            for (int c = 0; c < totalQIs; c++)
                if (sensitiveFlag[c] == 0)
                    generalizedQIsCount++;


            PointPairList[] pairEntropy = new PointPairList[generalizedQIsCount];

            for (int d = 0; d < pairEntropy.Length; d++)
                pairEntropy[d] = new PointPairList();

            ArrayList QI_Levels;
            int QIs = 0;

            Random random_RGB = new Random(DateTime.Now.Millisecond);
            Color[] color = new Color[generalizedQIsCount];

            int symbolIndex = 0;
            for (int a = 0; a < totalQIs; a++)
            {
                if (sensitiveFlag[a] == 1)
                    continue;

                QI_Levels = QIs_Levels[a];
                QIs = QI_Levels.Count;

                for (int b = 0; b < QI_Levels.Count; b++)
                {
                    tempLevel = (Level)QI_Levels[b];

                    if (tempLevel.levelEntropy > -0.5)
                        pairEntropy[pairCount].Add(QIs, tempLevel.levelEntropy);
                    else
                        pairEntropy[pairCount].Add(QIs, -0.5);
                    
                    QIs--;

                }//for (int b = 0; b < QIs_Levels[a].Count; b++)

                color[pairCount] = Color.FromArgb(255, random_RGB.Next(0, 255), random_RGB.Next(0, 255), random_RGB.Next(0, 255));
                                
                for (int d = 0; d < color.Length; d++)
                {
                    if ((d != pairCount) && (color[pairCount] == color[d]))
                    {
                        color[pairCount] = Color.FromArgb(255, random_RGB.Next(0, 255), random_RGB.Next(0, 255), random_RGB.Next(0, 255));
                        d = 0;
                    }
                }

                symbolIndex++;
                LineItem curveClassGain = treeGraphPane.AddCurve(qiNames[a], pairEntropy[pairCount], color[pairCount], RandomlySelectSymbol(ref symbolIndex));

                pairCount++;

            }//for (int a = 0; a < totalQIs; a++)


            treeGraph.AxisChange();

        }//public void CreateGraphEntropy(ZedGraphControl treeGraph)

        public void CreateGraphASpecLoss(ZedGraphControl treeGraph)
        {
            Level tempLevel;
            GraphPane treeGraphPane = treeGraph.GraphPane;



            //titles of the graph
            treeGraphPane.Title.Text = "Level Specification Loss";
            treeGraphPane.XAxis.Title.Text = "Levels";
            treeGraphPane.YAxis.Title.Text = "Specification Loss";

            treeGraphPane.XAxis.Scale.IsReverse = true;
            //treeGraphPane.YAxis.Scale.IsReverse = true;

            int generalizedQIsCount = 0;
            int pairCount = 0;

            for (int c = 0; c < totalQIs; c++)
                if (sensitiveFlag[c] == 0)
                    generalizedQIsCount++;


            PointPairList[] pairSpecLoss = new PointPairList[generalizedQIsCount];

            for (int d = 0; d < pairSpecLoss.Length; d++)
                pairSpecLoss[d] = new PointPairList();


            ArrayList QI_Levels;
            int QIs = 0;

            Random random_RGB = new Random(DateTime.Now.Millisecond);
            Color[] color = new Color[generalizedQIsCount];

            int symbolIndex = 0;
            for (int a = 0; a < totalQIs; a++)
            {
                if (sensitiveFlag[a] == 1)
                    continue;

                QI_Levels = QIs_Levels[a];

                QIs = QI_Levels.Count;

                for (int b = 0; b < QI_Levels.Count; b++)
                {
                    tempLevel = (Level)QI_Levels[b];

                    if (tempLevel.levelSpecLoss > -0.5)
                        pairSpecLoss[pairCount].Add(QIs, tempLevel.levelSpecLoss);
                    else
                        pairSpecLoss[pairCount].Add(QIs, -0.5);

                    QIs--;

                }//for (int b = 0; b < QIs_Levels[a].Count; b++)

                color[pairCount] = Color.FromArgb(255, random_RGB.Next(0, 255), random_RGB.Next(0, 255), random_RGB.Next(0, 255));
                                
                for (int d = 0; d < color.Length; d++)
                {
                    if ((d != pairCount) && (color[pairCount] == color[d]))
                    {
                        color[pairCount] = Color.FromArgb(255, random_RGB.Next(0, 255), random_RGB.Next(0, 255), random_RGB.Next(0, 255));
                        d = 0;
                    }
                }

                symbolIndex++;
                LineItem curveClassGain = treeGraphPane.AddCurve(qiNames[a], pairSpecLoss[pairCount], color[pairCount], RandomlySelectSymbol(ref symbolIndex));

                pairCount++;

            }//for (int a = 0; a < totalQIs; a++)

            treeGraph.AxisChange();

        }//public void CreateGraphSpecLoss(ZedGraphControl treeGraph)

        public void CreateGraphHistogram(ZedGraphControl treeGraph)
        {
            PointPairList pairHistogram = new PointPairList();
            GraphPane treeGraphPane = treeGraph.GraphPane;


            //titles of the graph
            treeGraphPane.Title.Text = "Frequency Distribution of Distinct Tuples";
            treeGraphPane.XAxis.Title.Text = "Distinct Tuple Index";
            treeGraphPane.YAxis.Title.Text = "Frequency";

            //this.logValues.WriteLine("Frequency Distribution of Distinct Tuples");
            

            DistinctTuple tempDistinctTuple;

            for (int a = 0; a < generalizedDistinctTuples.Count; a++)
            {
                tempDistinctTuple = (DistinctTuple)generalizedDistinctTuples[a];
                pairHistogram.Add(a, tempDistinctTuple.count);
                //this.logValues.WriteLine(tempDistinctTuple.count);
            }

            this.logValues.WriteLine("");

            //add all pair list in graph
            treeGraphPane.AddCurve("Distinct Tuple Frequency", pairHistogram, Color.Red, SymbolType.XCross);

            treeGraph.AxisChange();

        }//public void CreateGraphHistogram(ZedGraphControl zedGraphHistogram)


        public void CreateGraphClassFrequency(ZedGraphControl treeGraph)
        {
            PointPairList pairClassFrequency = new PointPairList();
            GraphPane treeGraphPane = treeGraph.GraphPane;

            //titles of the graph
            treeGraphPane.Title.Text = "Frequency of Different Equivalence Classes";
            treeGraphPane.XAxis.Title.Text = "Idex of Equivalence Classes";
            treeGraphPane.YAxis.Title.Text = "Frequency";

            int tempClassCount = 0;
            for (int a = 1; a < class_count.Length; a++)
            {
                for (int b = a; b < class_count.Length; b++)
                {
                    if (class_count[a] < class_count[b])
                    {
                        tempClassCount = class_count[a];
                        class_count[a] = class_count[b];
                        class_count[b] = tempClassCount;

                    }//if (class_count[a] > class_count[b])

                }// for (int b = 0; b < a; b++)

            }//for (int a = 1; a < class_count.Length; a++)


            for (int a = 0; a < class_count.Length; a++)
                if (class_count[a] > 0)
                    pairClassFrequency.Add(a, class_count[a]);

                //if (class_count_non_sorted[a] > 0)
                //    pairClassFrequency.Add(a, Math.Log(class_count_non_sorted[a]));

            //add all pair list in graph
            treeGraphPane.AddCurve("Class Size", pairClassFrequency, Color.Blue, SymbolType.Diamond);

            treeGraph.AxisChange();

        }//public void CreateGraphClassFrequency(ZedGraphControl treeGraph)

        public void CreateGraphDistributionDistance(ZedGraphControl treeGraph)
        {
            //Age               Magenta 
            //workclass         Aqua
            //Education Num     LimeGreen
            //Marital Status    HotPink
            //Occuption         Maroon
            //Race              Green
            //Sex               Red
            //Native Country    Blue

            int generalizedCount = 0;
            SuppressedClass suppressedClass;
            ArrayList tempClasses;

            for (int c = 0; c < totalQIs; c++)
                if (sensitiveFlag[c] == 0)
                    generalizedCount++;

            PointPairList[] pairQIsDistributionDistance = new PointPairList[generalizedCount];

            Random random_RGB = new Random();
            Color color;
           

            for (int d = 0; d < pairQIsDistributionDistance.Length; d++)
                pairQIsDistributionDistance[d] = new PointPairList();

            GraphPane treeGraphPane = treeGraph.GraphPane;

            //titles of the graph
            treeGraphPane.Title.Text = "Normalized Distribution Distance";
            treeGraphPane.XAxis.Title.Text = "Equivalence Class Size";
            treeGraphPane.YAxis.Title.Text = "Normalized Distance";

            generalizedCount = 0;
            for (int b = 0; b < totalQIs; b++)
                if (sensitiveFlag[b] == 1)
                    continue;
                else
                {
                    if (suppressedClasses_QIs == null)
                        continue;

                    tempClasses = (ArrayList)suppressedClasses_QIs[b];
                    
                    for (int a = 1; a < tempClasses.Count; a++)
                    {
                        suppressedClass = (SuppressedClass) tempClasses[a];
                        pairQIsDistributionDistance[generalizedCount].Add(a, suppressedClass.distribution_distance_normalized);
                    }
                    generalizedCount++;
                }

            int index = 0;
            for (int b = 0; b < totalQIs; b++)
            {
                if (sensitiveFlag[b] == 1)
                    continue;
                else
                {
                    color = Color.FromArgb(255, random_RGB.Next(0, 255), random_RGB.Next(0, 255), random_RGB.Next(0, 255));

                    treeGraphPane.AddCurve(qiNames[b], pairQIsDistributionDistance[index++], color, SymbolType.Star);
                }
            }   

            treeGraph.AxisChange();

        }//public void CreateGraphDistributionDistance(ZedGraphControl treeGraph)

        public void CreateGraphInfoGainRatio(ZedGraphControl treeGraph)
        {
            //Age               Magenta 
            //workclass         Aqua
            //Education Num     LimeGreen
            //Marital Status    HotPink
            //Occuption         Maroon
            //Race              Green
            //Sex               Red
            //Native Country    Blue


            int generalizedCount = 0;
            SuppressedClass suppressedClass;
            ArrayList tempClasses;

            for (int c = 0; c < totalQIs; c++)
                if (sensitiveFlag[c] == 0)
                    generalizedCount++;

            PointPairList[] pairQIsGainRation = new PointPairList[generalizedCount];

            for (int d = 0; d < pairQIsGainRation.Length; d++)
                pairQIsGainRation[d] = new PointPairList();



            GraphPane treeGraphPane = treeGraph.GraphPane;

            //titles of the graph
            treeGraphPane.Title.Text = "Class Information Gain Ratio";
            treeGraphPane.XAxis.Title.Text = "Class";
            treeGraphPane.YAxis.Title.Text = "Information Gain Ratio";

            Random random_RGB = new Random();
            Color color;
            
            generalizedCount = 0;
            for (int b = 0; b < totalQIs; b++)
                if (sensitiveFlag[b] == 1)
                    continue;
                else
                {
                    if (suppressedClasses_QIs == null)
                        continue;
                    
                    tempClasses = (ArrayList)suppressedClasses_QIs[b];

                    for (int a = 1; a < tempClasses.Count; a++)
                    {
                        suppressedClass = (SuppressedClass)tempClasses[a];
                        pairQIsGainRation[generalizedCount].Add(a, suppressedClass.distribution_distance_normalized);
                    }
                    generalizedCount++;
                }

            int index = 0;
            for (int b = 0; b < totalQIs; b++)
                if (sensitiveFlag[b] == 1)
                    continue;
                else
                {
                    color = Color.FromArgb(255, random_RGB.Next(0, 255), random_RGB.Next(0, 255), random_RGB.Next(0, 255));

                    treeGraphPane.AddCurve(qiNames[b], pairQIsGainRation[index++], color, SymbolType.Star);
                }

            treeGraph.AxisChange();


        }//public void CreateGraphInfoGainRatio(ZedGraphControl treeGraph)

        public void CreateGraphClassSizeXfrequency(ZedGraphControl treeGraph)
        {
            PointPairList pairHistogram = new PointPairList();
            GraphPane treeGraphPane = treeGraph.GraphPane;


            //titles of the graph
            treeGraphPane.Title.Text = "Tuple Counts of Different Equivalence Classes";
            treeGraphPane.XAxis.Title.Text = "Idex of Equivalence Classes";
            treeGraphPane.YAxis.Title.Text = "Tuples";

            ArrayList temp_class;
            DistinctTuple temp_tuple;

            logValues.WriteLine("");
            logValues.WriteLine("\nNumber of Tuples in each class after sorting");
            logValues.WriteLine("");
            Int64 totalTuplesSuppressed = 0;

            for (int a = 1; a <= class_count_tuples.Count; a++)
            {
                temp_class = (ArrayList)class_count_tuples[a];
                temp_tuple = (DistinctTuple)temp_class[0];

                pairHistogram.Add(a, temp_class.Count * temp_tuple.count);
                totalTuplesSuppressed += (temp_class.Count * temp_tuple.count);
                
                logValues.WriteLine("Class: " + a + ", (Distinct Tuples) size = " + temp_tuple.count + ", Frequency: " + temp_class.Count + ", (K = " + (temp_tuple.count + 1) + "), Tuples Suppressed in Class: " + (temp_class.Count * temp_tuple.count) + ", Total Tuples Suppressed: " + totalTuplesSuppressed);
                

            }

            this.logValues.WriteLine("===============================================================================");

            //add all pair list in graph
            treeGraphPane.AddBar("Class Tuples", pairHistogram, Color.Blue);

            treeGraph.AxisChange();

        }//public void CreateGraphClassSizeXfrequency(ZedGraphControl treeGraph)

        public void CreateGraphSuppressionCount(ZedGraphControl treeGraph)
        {
            PointPairList pairHistogram = new PointPairList();
            GraphPane treeGraphPane = treeGraph.GraphPane;
                        
            //titles of the graph
            treeGraphPane.Title.Text = "Tuples Suppression Count";
            treeGraphPane.XAxis.Title.Text = "Size of Equivalence Classes";
            treeGraphPane.YAxis.Title.Text = "Tuples";

            ArrayList temp_class;
            DistinctTuple temp_tuple;
            Int64 totalTuplesSuppressed = 0;

            for (int a = 1; a <= class_count_tuples.Count; a++)
            {
                temp_class = (ArrayList)class_count_tuples[a];
                temp_tuple = (DistinctTuple)temp_class[0];

                totalTuplesSuppressed += (temp_class.Count * temp_tuple.count);
                logValues.WriteLine("Class: " + a + ", (Distinct Tuples) size = " + temp_tuple.count + ", Frequency: " + temp_class.Count + ", (K = " + (temp_tuple.count + 1) + "), Tuples Suppressed in Class: " + (temp_class.Count * temp_tuple.count) + ", Total Tuples Suppressed: " + totalTuplesSuppressed);
                pairHistogram.Add(a, totalTuplesSuppressed);
            }


            //add all pair list in graph
            treeGraphPane.AddBar("Suppression Count", pairHistogram, Color.Blue);

            treeGraph.AxisChange();



        }//public void CreateGraphSuppressionCount(ZedGraphControl treeGraph)
       
        private void frmWekaAnonymize_Load(object sender, EventArgs e)
        {
            
            
           // MessageBox.Show("Done");
        }


    }//public partial class frmWekaAnonymize : Form

}//namespace WekaAnonymize

